import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)

known_emulators = []
default_emulator = None

debug_import = False

try:
    from .atari8bit import wxAtari800, wxAtari800XL, wxAtari5200
    known_emulators.append(wxAtari800)
    known_emulators.append(wxAtari800XL)
    known_emulators.append(wxAtari5200)
    if default_emulator is None:
        default_emulator = wxAtari800
except ImportError as e:
    log.warning(f"Atari800 emulator not available: {e}")
    if debug_import:
        raise

try:
    from .generic6502 import Generic6502
    known_emulators.append(Generic6502)
    if default_emulator is None:
        default_emulator = Generic6502
except ImportError as e:
    log.warning(f"Generic6502 emulator not available: {e}")
    if debug_import:
        raise

try:
    from .apple2 import Crabapple
    known_emulators.append(Crabapple)
    if default_emulator is None:
        default_emulator = Crabapple
except ImportError as e:
    log.warning(f"Crabapple emulator not available: {e}")
    if debug_import:
        raise



import sys

import wx

# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory
from apptools.preferences.api import PreferencesHelper
from traits.api import Bool, Dict, Enum, List, Str, Unicode, Int, Font, Range, Tuple, Color, Any, Property, cached_property
from traitsui.api import FontEditor, HGroup, VGroup, Item, Label, \
    View, RangeEditor, ColorEditor

if sys.platform == "darwin":
    def_font = "10 point Monaco"
elif sys.platform == "win32":
    def_font = "10 point Lucida Console"
else:
    def_font = "10 point monospace"


def byte2hex(val):
    return "$02x" % val

def word2hex(val):
    return "$04x" % val

def int2hex(val):
    return "%x" % val

byte2str = byte2hex
word2str = word2hex
int2str = int2hex

class EmulatorPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = 'sawx.task.emulator'

    #### Preferences ##########################################################

    map_width_low = 1
    map_width_high = 256
    map_width = Range(low=map_width_low, high=map_width_high, value=40)

    bitmap_width_low = 1
    bitmap_width_high = 16
    bitmap_width = Range(low=bitmap_width_low, high=bitmap_width_high, value=1)

    hex_grid_width_low = 1
    hex_grid_width_high = 256
    hex_grid_width = Range(low=hex_grid_width_low, high=hex_grid_width_high, value=16)

    # Font used for hex/disassembly
    text_font = Font(def_font)

    text_font_char_width = Property(Int, depends_on='text_font')
  
    text_font_char_height = Property(Int, depends_on='text_font')

    image_caches = Property(Dict, depends_on='text_font')

    header_font = Font(def_font + " bold")

    hex_grid_lower_case = Bool(True)

    assembly_lower_case = Bool(False)

    int_display_format = Enum(
        "Hexadecimal",
        "Decimal",
        "Both",
    )

    hex_display_format = Enum(
        "$XX",
        "0xXX",
        "XXh",
    )

    disassembly_column_widths = Tuple(0, 0, 0)

    background_color = Color(wx.WHITE)

    text_color = Color(wx.BLACK)

    highlight_background_color = Color(wx.Colour(100, 200, 230))

    data_background_color = Color(wx.Colour(224, 224, 224))

    empty_background_color = Color(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).Get(False))

    match_background_color = Color(wx.Colour(255, 255, 180))

    comment_background_color = Color(wx.Colour(255, 180, 200))

    error_background_color = Color(wx.Colour(255, 128, 128))

    diff_text_color = Color(wx.Colour(255, 0, 0))

    unfocused_caret_color = Color(wx.Colour(128, 128, 128))
    
    row_header_bg_color = Color(wx.Colour(224, 224, 224))
    
    col_header_bg_color = Color(wx.Colour(224, 224, 224))

    col_label_border_width = Int(3)

    row_label_border_width = Int(3)

    row_height_extra_padding = Int(-3)

    base_cell_width_in_chars = Int(2)

    pixel_width_padding = Int(2)
    
    caret_pen = Any

    selected_brush = Any
    
    normal_brush = Any
    
    data_brush = Any
    
    match_brush = Any
    
    comment_brush = Any
    
    empty_brush = Any

    def _caret_pen_default(self):
        return wx.Pen(self.unfocused_caret_color, 1, wx.SOLID)

    def _selected_brush_default(self):
        return wx.Brush(self.highlight_background_color, wx.SOLID)

    def _normal_brush_default(self):
        return wx.Brush(self.background_color, wx.SOLID)

    def _data_brush_default(self):
        return wx.Brush(self.data_background_color, wx.SOLID)

    def _match_brush_default(self):
        return wx.Brush(self.match_background_color, wx.SOLID)

    def _comment_brush_default(self):
        return wx.Brush(self.comment_background_color, wx.SOLID)

    def _empty_brush_default(self):
        return wx.Brush(self.empty_background_color, wx.SOLID)

    @cached_property
    def _get_text_font_char_width(self):
        dc = wx.MemoryDC()
        dc.SetFont(self.text_font)
        return dc.GetCharWidth()

    @cached_property
    def _get_text_font_char_height(self):
        dc = wx.MemoryDC()
        dc.SetFont(self.text_font)
        return dc.GetCharHeight()

    @cached_property
    def _get_image_caches(self):
        return dict()

    def calc_cell_size_in_pixels(self, chars_per_cell):
        width = self.pixel_width_padding * 2 + self.text_font_char_width * chars_per_cell
        height = self.row_height_extra_padding + self.text_font_char_height
        return width, height

    def calc_text_width(self, text):
        return self.text_font_char_width * len(text)

    def calc_image_cache(self, cache_cls):
        try:
            c = self.image_caches[cache_cls]
        except KeyError:
            c = cache_cls(self)
            self.image_caches[cache_cls] = c
        return c

    @property
    def hex_format_character(self):
        return "x" if self.hex_grid_lower_case else "X"


class EmulatorPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = EmulatorPreferences

    category = Str('Editors')

    #### 'FrameworkPreferencesPane' interface ################################

    # Note the quirk in the RangeEditor: specifying a custom editor is
    # supposed to take the defaults from the item name specified, but I
    # can't get it to work with only the "mode" parameter.  I have to specify
    # all the other params, and the low/high values have to be attributes
    # in EmulatorPreferences, not the values in the trait itself.  See
    # traitsui/editors/range_editor.py
    view = View(
        VGroup(HGroup(Item('map_width', editor=RangeEditor(mode="spinner", is_float=False, low_name='map_width_low', high_name='map_width_high')),
                      Label('Default Character Map Width (in bytes)'),
                      show_labels = False),
               HGroup(Item('bitmap_width', editor=RangeEditor(mode="spinner", is_float=False, low_name='bitmap_width_low', high_name='bitmap_width_high')),
                      Label('Default Bitmap Width (in bytes)'),
                      show_labels = False),
               HGroup(Item('hex_grid_width', editor=RangeEditor(mode="spinner", is_float=False, low_name='hex_grid_width_low', high_name='hex_grid_width_high')),
                      Label('Default Hex Grid Width (in bytes)'),
                      show_labels = False),
               HGroup(Item('text_font'),
                      Label('Hex Display Font'),
                      show_labels = False),
               HGroup(Item('header_font'),
                      Label('Column Header Font'),
                      show_labels = False),
               HGroup(Item('int_display_format'),
                      Label('Number Display Format'),
                      show_labels=False),
               HGroup(Item('hex_display_format'),
                      Label('Hex Display Format'),
                      show_labels=False),
               HGroup(Item('hex_grid_lower_case'),
                      Label('Use Lower Case for Hex Digits'),
                      show_labels = False),
               HGroup(Item('assembly_lower_case'),
                      Label('Use Lower Case for Assembler Mnemonics'),
                      show_labels = False),
               HGroup(Item('highlight_background_color', editor=ColorEditor(), style='custom'),
                      Label('Highlight Color'),
                      show_labels = False),
               label='Hex Editor'),
        resizable=True)



import os
import tempfile

import numpy as np

from atrip import find_diskimage

from ..debugger import Debugger
from ..debugger.dtypes import FRAME_STATUS_DTYPE
from .save_state import FrameHistory
from .. import disassembler as disasm
from ..utils.archutil import Labels, load_memory_map

import logging
log = logging.getLogger(__name__)


# Values must correspond to values in libdebugger.h
FRAME_START = 0
FRAME_FINISHED = 1
FRAME_BREAKPOINT = 2
FRAME_WATCHPOINT = 3


class EmulatorBase(Debugger):
    cpu = "<base>"
    ui_name = "<pretty name>"

    mime_prefix = ""
    mime_types = set()

    serializable_attributes = ['input_raw', 'output_raw', 'frame_count']
    serializable_computed = {'input_raw', 'output_raw'}

    input_array_dtype = None
    output_array_dtype = None
    width = 320
    height = 192

    low_level_interface = None  # cython module; e.g.: libatari800, lib6502

    history_entry_dtype = disasm.HISTORY_ENTRY_DTYPE

    def __init__(self):
        Debugger.__init__(self)
        self.input_raw = np.zeros([self.input_array_dtype.itemsize], dtype=np.uint8)
        self.input = self.input_raw.view(dtype=self.input_array_dtype)
        self.output_raw = np.zeros([FRAME_STATUS_DTYPE.itemsize + self.output_array_dtype.itemsize], dtype=np.uint8)
        self.status = self.output_raw[0:FRAME_STATUS_DTYPE.itemsize].view(dtype=FRAME_STATUS_DTYPE)
        self.output = self.output_raw[FRAME_STATUS_DTYPE.itemsize:].view(dtype=self.output_array_dtype)
        self.bootfile = None
        self.frame_count = 0
        self.frame_event = []
        self.frame_history = FrameHistory()
        self.offsets = None
        self.names = None
        self.save_state_memory_blocks = None
        self.main_memory = None
        self.last_boot_state = None
        self.forced_modifier = None
        self.emulator_started = False
        self.cpu_history = None
        self.labels = None

        self.compute_color_map()
        self.screen_rgb, self.screen_rgba = self.calc_screens()

    @property
    def raw_array(self):
        return self.output_raw

    @property
    def raw_state(self):
        return self.output_raw[self.state_start_offset:]

    @property
    def video_array(self):
        return self.output['video'][0]

    @property
    def audio_array(self):
        return self.output['audio'][0]

    @property
    def state_array(self):
        return self.output['state'][0]

    @property
    def memory_access_array(self):
        return self.status['memory_access'][0]

    @property
    def access_type_array(self):
        return self.status['access_type'][0]

    @property
    def current_frame_number(self):
        return self.status['frame_number'][0]

    @property
    def is_frame_finished(self):
        return self.status['frame_status'][0] == FRAME_FINISHED

    @property
    def current_cycle_in_frame(self):
        return self.status['current_cycle_in_frame'][0]

    @property
    def cycles_user(self):
        return self.status['cycles_user'][0]

    @property
    def cycles_since_power_on(self):
        return self.status['cycles_since_power_on'][0]

    @property
    def instructions_since_power_on(self):
        return self.status['instructions_since_power_on'][0]

    @property
    def stack_pointer(self):
        raise NotImplementedError("define stack_pointer property in subclass")

    @stack_pointer.setter
    def stack_pointer(self, value):
        raise NotImplementedError("define stack_pointer property in subclass")

    @property
    def program_counter(self):
        raise NotImplementedError("define program_counter property in subclass")

    @program_counter.setter
    def program_counter(self, value):
        raise NotImplementedError("define program_counter property in subclass")

    @property
    def current_cpu_status(self):
        return "not running"

    @property
    def has_save_points(self):
        return len(self.frame_history) > 0

    @classmethod
    def guess_from_document(cls, document):
        try:
            mime = document.metadata.mime
        except:
            pass
        else:
            if mime in cls.mime_types:
                return True
            elif cls.mime_prefix and mime.startswith(cls.mime_prefix):
                return True
        return False

    ##### Video

    def compute_color_map(self):
        pass

    def calc_screens(self):
        rgb = np.empty((self.height, self.width, 3), np.uint8)
        rgba = np.empty((self.height, self.width, 4), np.uint8)
        return rgb, rgba

    ##### Serialization

    def restore_computed_attributes(self, state):
        Debugger.restore_computed_attributes(self, state)
        self.input_raw[:] = state['input_raw']
        self.restore_state(state['output_raw'])

        # recalculate any internal offsets within output_raw, e.g. atari800
        # needs this because the size of the state save file changes depending
        # on the emulator configuration
        self.low_level_interface.get_current_state(state['output_raw'])
        self.configure_save_state_memory_blocks()

    ##### Initialization

    def configure_emulator(self, emu_args=None, instruction_history_count=100000, *args, **kwargs):
        self.configure_labels()
        self.init_cpu_history(instruction_history_count)
        self.args = self.process_args(emu_args)
        self.low_level_interface.clear_state_arrays(self.input, self.output_raw)
        self.low_level_interface.start_emulator(self.args)
        self.configure_emulator_defaults()
        self.emulator_started = True
        self.configure_io_arrays()

    def configure_labels(self, labels=None):
        if labels is None:
            labels = load_memory_map(self.name)
        self.labels = labels

    def configure_emulator_defaults(self):
        pass

    def configure_io_arrays(self):
        if not self.emulator_started:
            print(f"configure_io_arrays can't be called before emulator started")
            return
        print(f"input_raw: {self.input_raw.shape}  {self.input_raw}")
        print(f"output_raw: {self.output_raw.shape} {self.output_raw}")
        self.low_level_interface.configure_state_arrays(self.input, self.output_raw)
        self.configure_save_state_memory_blocks()

    def configure_save_state_memory_blocks(self):
        self.parse_state()
        self.generate_save_state_memory_blocks()
        self.cpu_state = self.calc_cpu_data_array()
        self.main_memory = self.calc_main_memory_array()

    def process_args(self, emu_args):
        return emu_args if emu_args else []

    def add_segment_to_memory(self, segment):
        start = segment.origin
        end = (start + len(segment)) & 0xffff
        count = end - start
        log.debug(f"Copying {segment} to memory: {start:#04x}-{end:#04x}")
        self.main_memory[start:end] = segment.data[:count]

    ##### Machine boot

    def find_default_boot_segment(self, segments):
        for segment in segments:
            if segment.origin > 0:
                return segment
        return None

    def boot_from_segment(self, boot_segment):
        if boot_segment is not None:
            data = boot_segment.data
            origin = boot_segment.origin
            self.boot_from_raw(data, origin)

    def boot_from_raw(self, data, origin):
        if self.bootfile is not None:
            try:
                os.remove(self.bootfile)
                self.bootfile = None
            except:  # MSW raises WindowsError, but that's not defined cross-platform
                log.warning("Unable to remove temporary boot file %s." % self.bootfile)
        fd, self.bootfile = tempfile.mkstemp(".omnivore_boot_segment")
        fh = os.fdopen(fd, "wb")
        fh.write(data)
        fh.close()
        log.debug(f"Created temporary file {self.bootfile} to use as boot disk image")
        self.boot_from_file(self.bootfile)

    def boot_from_file(self, filename):
        parser = find_diskimage(filename, True)
        print(f"diskimage: filename={filename} image={parser.image}")
        run_addr = None
        for s in parser.image.segments:
            print(f"segment: {s}")
            try:
                run_addr = s.run_address()
            except AttributeError:
                if run_addr is None:
                    run_addr = s.origin
            self.add_segment_to_memory(s)
        print(f"running at: {hex(run_addr)}")
        self.program_counter = run_addr
        self.last_boot_state = self.calc_current_state()
        self.coldstart()

    def parse_state(self):
        base = np.byte_bounds(self.output_raw)[0]
        self.state_start_offset = np.byte_bounds(self.state_array)[0] - base

        memaccess_offset = np.byte_bounds(self.memory_access_array)[0] - base
        memtype_offset = np.byte_bounds(self.access_type_array)[0] - base
        video_offset = np.byte_bounds(self.video_array)[0] - base
        audio_offset = np.byte_bounds(self.audio_array)[0] - base
        self.save_state_memory_blocks = [
            (memaccess_offset, self.memory_access_array.nbytes, 0, "Memory Access"),
            (memtype_offset, self.access_type_array.nbytes, 0, "Access Type"),
            (video_offset, self.video_array.nbytes, 0, "Video Frame"),
            (audio_offset, self.audio_array.nbytes, 0, "Audio Data"),
        ]

    def generate_save_state_memory_blocks(self):
        pass

    def next_frame(self):
        self.process_key_state()
        if not self.is_frame_finished:
            print(f"next_frame: continuing frame from cycle {self.current_cycle_in_frame} of frame {self.current_frame_number}")
        bpid = self.low_level_interface.next_frame(self.input, self.output_raw, self.debug_cmd, self.cpu_history)
        if self.is_frame_finished:
            self.frame_count += 1
            self.process_frame_events()
            self.save_history()
        self.forced_modifier = None
        return self.get_breakpoint(bpid)

    def process_frame_events(self):
        still_waiting = []
        for count, callback in self.frame_event:
            if self.frame_count >= count:
                log.debug("processing %s", callback)
                callback()
            else:
                still_waiting.append((count, callback))
        self.frame_event = still_waiting

    def end_emulation(self):
        pass

    def debug_video(self):
        """Return text based view of portion of video array, for debugging
        purposes only so it doesn't have to be fast.
        """
        pass

    def debug_state(self):
        """Show CPU status registers
        """
        print(self.current_cpu_status)

    # Emulator user input functions

    def coldstart(self):
        """Simulate an initial power-on startup.
        """
        pass

    def warmstart(self):
        """Simulate a warm start; i.e. pressing the system reset button
        """
        pass

    def keypress(self, ascii_char):
        """Pass an ascii char to the emulator
        """
        self.send_char(ord(ascii_char))

    def joystick(self, stick_num, direction_value, trigger_pressed=False):
        """Pass a joystick/trigger value to the emulator
        """
        pass

    def paddle(self, paddle_num, paddle_percentage):
        """Pass a paddle value to the emulator
        """
        pass

    def process_key_state(self):
        """Read keyboard and compute any values that should be sent to the
        emulator.
        """
        pass

    def process_key_down(self, evt, keycode):
        pass

    def process_key_up(self, evt, keycode):
        pass

    ##### Input routines

    def send_char(self, key_char):
        print(f"sending char: {key_char}")
        self.input['keychar'] = key_char
        self.input['keycode'] = 0
        self.input['special'] = 0

    def send_keycode(self, keycode):
        self.input['keychar'] = 0
        self.input['keycode'] = keycode
        self.input['special'] = 0

    def send_special_key(self, key_id):
        self.input['keychar'] = 0
        self.input['keycode'] = 0
        self.input['special'] = key_id

    def clear_keys(self):
        self.input['keychar'] = 0
        self.input['keycode'] = 0
        self.input['special'] = 0

    # Utility functions

    def load_disk(self, drive_num, pathname):
        self.low_level_interface.load_disk(drive_num, pathname)

    def calc_current_state(self):
        return self.output_raw.copy()

    def save_history(self, force=False):
        # History is saved in a big list, which will waste space for empty
        # entries but makes things extremely easy to manage. Simply delete
        # a history entry by setting it to NONE.
        frame_number = int(self.status['frame_number'][0])
        if force or self.frame_history.is_memorable(frame_number):
            print(f"Saving history at {frame_number}")
            d = self.calc_current_state()
            self.frame_history.save_frame(frame_number, d)
            self.print_history(frame_number)

    def get_history(self, frame_number):
        frame_number = int(frame_number)
        raw = self.frame_history[frame_number]
        status = raw[0:FRAME_STATUS_DTYPE.itemsize].view(dtype=FRAME_STATUS_DTYPE)
        output = raw[FRAME_STATUS_DTYPE.itemsize:].view(dtype=self.output_array_dtype)
        return status, output

    def restore_state(self, state):
        self.low_level_interface.restore_state(state)
        self.output_raw[:] = state

    def restore_history(self, frame_number):
        print(("restoring state from frame %d" % frame_number))
        frame_number = int(frame_number)
        if frame_number < 0:
            return
        try:
            d = self.frame_history[frame_number]
        except KeyError:
            log.error(f"{frame_number} not in history")
            pass
        else:
            self.restore_state(d)
            # self.frame_history[(frame_number + 1):] = []  # remove frames newer than this
            # print(("  %d items remain in history" % len(self.frame_history)))
            # self.frame_event = []

    def print_history(self, frame_number):
        d = self.frame_history[frame_number]
        status = d[:FRAME_STATUS_DTYPE.itemsize].view(dtype=FRAME_STATUS_DTYPE)
        output = d[FRAME_STATUS_DTYPE.itemsize:].view(dtype=self.output_array_dtype)
        print("history[%d] of %d: %d %s" % (status['frame_number'], len(self.frame_history), len(d), output['state'][0][0:8]))

    def get_previous_history(self, frame_cursor):
        return self.frame_history.get_previous_frame(frame_cursor)

    def get_next_history(self, frame_cursor):
        return self.frame_history.get_next_frame(frame_cursor)

    # graphics

    def get_color_indexed_screen(self, frame_number=-1):
        """Return color indexed screen in whatever native format this
        emulator supports
        """
        pass

    def get_frame_rgb(self, frame_number=-1):
        """Return RGB image of the current screen
        """

    def get_frame_rgba(self, frame_number=-1):
        """Return RGBA image of the current screen
        """

    def get_frame_rgba_opengl(self, frame_number=-1):
        """Return RGBA image of the current screen, suitable for use with
        OpenGL (flipped vertically)
        """

    # CPU history

    def init_cpu_history(self, num_entries):
        self.cpu_history = disasm.HistoryStorage(num_entries)

    def cpu_history_show_range(self, from_index, details=False):
        self.cpu_history.debug_range(from_index)

    def cpu_history_show_next_instruction(self):
        self.low_level_interface.show_next_instruction(self.cpu_history)

    def calc_stringified_history(self, start_index, count):
        """Returns an list of indexes into the entries array for the
        range of history requested.

        Because it's a circular buffer, the wraparound case is unable to be
        handled by numpy, so instead is used to map the row number in the
        display window (which ranges from 0 -> count) to the history entry
        starting at first_entry_index + start_index for count entries.
        """
        return self.cpu_history.stringify(start_index, count, self.labels.labels)

    @property
    def num_cpu_history_entries(self):
        return len(self.cpu_history)



""" Action definitions for emulators

"""
import os
import sys

import wx
import wx.lib.dialogs

# Enthought library imports.
from traits.api import on_trait_change, Any, Int, Bool
from pyface.api import YES, NO

from sawx.framework.enthought_api import Action, ActionItem, EditorAction, NameChangeAction, TaskDynamicSubmenuGroup
from sawx.utils.command import StatusFlags
from sawx.framework import loader

from subprocess import *
from sawx.ui.dialogs import prompt_for_hex, prompt_for_dec, prompt_for_string, get_file_dialog_wildcard, ListReorderDialog
from ..ui.dialogs import SegmentOrderDialog
from .document import EmulationDocument

if sys.platform == "darwin":
    RADIO_STYLE = "toggle"
else:
    RADIO_STYLE = "radio"

import logging
log = logging.getLogger(__name__)


class UseEmulatorAction(EditorAction):
    """Change the default emulator
    """
    # Traits
    name = "<emu>"
    emulator = Any
    style = RADIO_STYLE

    def perform(self, event):
        self.active_editor.document.emulator_class_override = self.emulator


class BootDiskImageAction(EditorAction):
    """Run the current disk image in an emulator
    """
    name = "Boot Current Disk Image"
    tooltip = "Start emulator using the current file as the boot disk"

    def perform(self, event=None):
        source = self.active_editor.document
        doc = EmulationDocument.create_document(source_document=source, emulator_type=source.emulator_class_override)
        try:
            doc.boot()
        except RuntimeError as e:
            self.task.error(str(e))
        else:
            loader.load_document(doc, in_current_window=False)


class BootSegmentAction(EditorAction):
    """Start an emulator by pre-populating memory using the contents of some
    selected segments.
    """
    name = "Boot Current Segment"
    tooltip = "Start emulator using current segment to pre-fill the memory of the emulator"

    def perform(self, event=None):
        source = self.active_editor.document
        doc = EmulationDocument.create_document(source_document=source, emulator_type=source.emulator_class_override)
        try:
            doc.boot(self.active_editor.segment)
        except RuntimeError as e:
            self.task.error(str(e))
        else:
            loader.load_document(doc, in_current_window=False)


class LoadSegmentAction(EditorAction):
    """Initialize memory of an emulator using the contents of some
    selected segments.
    """
    name = "Load Current Segment"
    tooltip = "Create (but not boot) an emulator using current segment to pre-fill the memory"

    def perform(self, event=None):
        source = self.active_editor.document
        doc = EmulationDocument(source_document=source, emulator_type=source.emulator_class_override)
        doc.load(self.active_editor.segment)
        loader.load_document(doc)


class EmulatorAction(EditorAction):
    """Base class for emulator actions
    """
    name = "<emulator action>"
    tooltip = "control the emulator"

    def perform(self, event=None):
        print("emulate!")


class PauseResumeAction(EmulatorAction):
    """Stop/Restart the emulation
    """
    name = "Resume"
    tooltip = "Pause or restart the emulation"
    accelerator = 'F8'

    def perform(self, event=None):
        if self.active_editor.document.emulator_running:
            self.active_editor.document.pause_emulator()
        else:
            self.active_editor.document.restart_emulator()

    def _update_enabled(self, ui_state):
        if not self.active_editor.document.emulator_running:
            self.name = "Resume"
        else:
            self.name = "Pause"


class StepAction(EmulatorAction):
    """Restart the emulation
    """
    name = "Step"
    tooltip = "Restart the emulation"
    accelerator = 'F9'

    def perform(self, event=None):
        self.active_editor.document.debugger_step()

    def _update_enabled(self, ui_state):
        self.enabled = self.active_editor.document.emulator_paused


class BreakVBIStart(EmulatorAction):
    """Break at the start of the next VBI
    """
    name = "Break at Next VBI Start"
    tooltip = "Enter debugger at the beginning of the next VBI"
    accelerator = 'shift-F9'

    def perform(self, event=None):
        self.active_editor.document.debugger_break_vbi_start()

    def _update_enabled(self, ui_state):
        self.enabled = self.active_editor.document.emulator_paused


class StepIntoAction(StepAction):
    """Restart the emulation
    """
    name = "Step Into"
    tooltip = "Restart the emulation"
    accelerator = 'F10'

    def perform(self, event=None):
        print("resume!")


class StepOverAction(StepAction):
    """Restart the emulation
    """
    name = "Step Over"
    tooltip = "Restart the emulation"
    accelerator = 'F11'

    def perform(self, event=None):
        print("resume!")


class EndOfFrameAction(StepAction):
    """Restart the emulation
    """
    name = "Run to End of Frame"
    tooltip = "Break when reaching the end of the current frame"
    accelerator = 'F12'

    def perform(self, event=None):
        self.active_editor.document.debugger_count_frames()


class StartAction(EmulatorAction):
    name = "Start"
    tooltip = "Press the Start button"
    accelerator = 'F4'

    def perform(self, event=None):
        print("START!")
        self.active_editor.document.emulator.forced_modifier = "start"


class SelectAction(EmulatorAction):
    name = "Select"
    tooltip = "Press the Select button"
    accelerator = 'F3'

    def perform(self, event=None):
        self.active_editor.document.emulator.forced_modifier = "select"


class OptionAction(EmulatorAction):
    name = "Option"
    tooltip = "Press the Option button"
    accelerator = 'F2'

    def perform(self, event=None):
        self.active_editor.document.emulator.forced_modifier = "option"


class ColdstartAction(EmulatorAction):
    name = "Reboot"
    tooltip = "Simulate turning off the power and turning it back on again"
    accelerator = 'Shift F5'

    def perform(self, event=None):
        self.active_editor.document.emulator.coldstart()


class WarmstartAction(EmulatorAction):
    name = "System Reset"
    tooltip = "Simulate pressing the system reset key"
    accelerator = 'F5'

    def perform(self, event=None):
        self.active_editor.document.emulator.warmstart()


class PreviousSaveStateAction(EmulatorAction):
    """Load the previous saved frame into the current state of the emulator
    """
    name = "Previous Save Point"
    tooltip = "Show previous frame in saved state of the history"
    accelerator = 'F6'

    def perform(self, event=None):
        d = self.active_editor.document
        if d.emulator_running:
            d.pause_emulator()
        else:
            d.history_previous()


class NextSaveStateAction(EmulatorAction):
    """Load the next saved frame into the current state of the emulator
    """
    name = "Next Save Point"
    tooltip = "Show next frame in saved state of the history"
    accelerator = 'F7'

    def perform(self, event=None):
        d = self.active_editor.document 
        print("OEUOEU", d.emulator_running)
        if d.emulator_running:
            d.pause_emulator()
        else:
            print("BBVJKBVJKBQJBKQB")
            d.history_next()



import re
import collections

import numpy as np

from sawx.persistence import get_template, iter_templates

from ..disassembler.libudis import LabelStorage

import logging
log = logging.getLogger(__name__)



line_re = re.compile("^([a-fA-F0-9]+)\s*([a-zA-Z][a-zA-Z0-9_]*)*\s*([a-zA-Z][a-zA-Z0-9_]*)?\s*(\S*)^")

type_codes = {
    'b': 0, # "byte",
    'w': 1, # "word",
    'l': 3, # "long",
    }

desc_codes = {
    'h': 0x00, # "hex",
    'hex': 0x00,
    'c': 0x10, # "char", "ascii",
    'char': 0x10,
    'ascii': 0x10,
    'a': 0x20, # "atascii",
    'atascii': 0x20,
    'd': 0x30, # "decimal",
    'decimal': 0x30,
    'b': 0x40, # "binary",
    'bin': 0x40,
    'binary': 0x40,
}

SourceLabel = collections.namedtuple('SourceLabel', 'label byte_count item_count type_code desc_code')


class Labels:
    @classmethod
    def from_file(cls, filename):
        text = open(filename).read()
        return cls.from_text(text)

    @classmethod
    def from_text(cls, text):
        m = cls()
        try:
            text = text.decode('utf-8')
        except AttributeError as e:
            # already a string!
            pass
        line_num = 0
        for line in text.splitlines():
            line = line.strip()
            line_num += 1
            if not line or line.startswith("#") or line.startswith(";"):
                continue
            line = line.split("#")[0]
            tokens = line.split()
            count = len(tokens)
            if count < 2:
                log.warning(f"Missing label on line {line_num}")
                continue
            try:
                addr = int(tokens[0], 16)
            except ValueError:
                log.warning(f"Invalid hex digits on line {line_num}: '{tokens[0]}'")
                continue
            if "/" in tokens[1]:
                read_val, write_val = tokens[1].split("/")
            else:
                read_val = tokens[1]
                write_val = None
            try:
                type_description = tokens[2]
            except IndexError:
                type_description = 'b'
            try:
                desc_code = desc_codes[tokens[3]]
            except IndexError:
                desc_code = desc_codes['hex']
            try:
                m.add(addr, read_val, write_val, type_description, desc_code)
            except RuntimeError:
                log.warning("Invalid type code on line {line_num}: `'{type_code}'")
        return m

    def __init__(self):
        self.labels = LabelStorage()
        self.write_labels = LabelStorage()

    def add(self, addr, r, w, type_description, desc_code):
        type_code = type_codes[type_description[-1]]
        try:
            item_count = int(type_description[:-1])
        except ValueError:
            item_count = 1
        num_bytes = item_count * (type_code + 1)
        label = SourceLabel(r.encode('utf-8'), num_bytes, item_count, type_code, desc_code)
        self.labels[addr] = label
        if w:
            # print(f"found write label {w} at {addr}")
            self.write_labels[addr] = w.encode('utf-8')
            # print(f"stored as: {self.get_name(addr, True)}")

    def get_name(cls, addr, write=False):
        if write:
            if addr in cls.write_labels:
                return cls.write_labels[addr][0].decode('utf-8')
        if addr in cls.labels:
            return cls.labels[addr][0].decode('utf-8')
        return ""

    def __contains__(self, addr):
        return addr in self.labels

    def __str__(self):
        lines = []
        a = sorted(set(self.labels.keys()).union(set(self.write_labels.keys())))
        for addr in a:
            try:
                s = self.labels[addr]
            except KeyError:
                s = SourceLabel(addr, "", 1, "b")
            r = s[0].decode('utf-8')
            w = self.get_name(addr, True)
            if r == w:
                w = ""
            lines.append(f"{addr:4x}: {r:8} {w:8} {s[3]}")
        return "\n".join(lines)

    def update(self, other):
        self.labels.update(other.labels)
        self.write_labels.update(other.write_labels)


machine_labels = {}

def load_memory_map(keyword):
    global machine_labels
    try:
        labels = machine_labels[keyword]
    except KeyError:
        try:
            text = get_template(keyword)
        except OSError as e:
            try:
                text = get_template(keyword + ".labels")
            except OSError as e:
                log.error(f"Couldn't find memory map named '{keyword}'")
                return Labels()
        labels = Labels.from_text(text)
        machine_labels[keyword] = labels
    return labels

available_memory_maps = {}

def calc_available_memory_maps():
    global available_memory_maps
    if not available_memory_maps:
        for template in iter_templates("labels"):
            available_memory_maps[template.keyword] = template
    return available_memory_maps



# Standard library imports.
import sys
import os

# Major package imports.
import wx
import numpy as np
import json

from sawx.editor import SawxEditor
from sawx.ui.tilemanager import TileManager
from sawx.persistence import get_template
from sawx.events import EventHandler

from ..viewer import find_viewer_class_by_name

import logging
log = logging.getLogger(__name__)


class DummyLinkedBase(object):
    segment = None
    segment_uuid = None

class DummyFocusedViewer(object):
    linked_base = DummyLinkedBase


class TileManagerBase(SawxEditor):
    """Base class for editors that use the tile manager as the central window
    manager

    The TileManager splits up the central window into a set of viewers that ar
    arranged in a layout in relation to one another. This layout can be saved
    and restored along with the data being edited.

    The layout can also be overridden when loading an editor. The layout/viewer
    creation priority is as follows:

    * viewers specified on the command line (any saved layout will be ignored)
    * viewers & layout restored from the save file
    * viewers & layout from a template file based on a combination of the
      editor and MIME type of the file
    * viewers specified in the default_viewers class attribute of the editor
    """
    editor_id = "omnivore.tile_manager_base"
    ui_name = "Tile Manager Base"

    default_viewers = "dummy"

    def __init__(self, document, action_factory_lookup=None):
        SawxEditor.__init__(self, document, action_factory_lookup)
        self.focused_viewer = None
        self.focused_viewer_changed_event = EventHandler(self)
        self.viewers = []

    def prepare_destroy(self):
        self.focused_viewer = None
        # Operate on copy of list because you can't iterate and remove from the
        # same list
        for v in list(self.viewers):
            log.debug(f"Closing viewer: {v}")
            self.viewers.remove(v)
            v.prepare_for_destroy()
            del v

    def create_control(self, parent):
        """Creates the `TileManager` control that will manage all the viewers
        """

        panel = TileManager(parent, toggle_checker=self.check_viewer_center_base)
        panel.Bind(TileManager.EVT_CLIENT_ACTIVATED, self.on_viewer_active)
        panel.Bind(TileManager.EVT_CLIENT_CLOSE, self.on_viewer_close)
        panel.Bind(TileManager.EVT_CLIENT_REPLACE, self.on_viewer_replace)
        panel.Bind(TileManager.EVT_CLIENT_TOGGLE_REQUESTED, self.on_viewer_link)
        return panel

    def has_command_line_viewer_override(self, args):
        return bool(args)

    def get_default_layout(self):
        template_name = self.document.calc_layout_template_name(self.editor_id)
        log.debug("template from: %s" % template_name)
        try:
            data = get_template(template_name)
        except OSError as err:
            log.error(f"Failed loading template: {err}")
            layout = {}
        else:
            try:
                layout = json.loads(data)
            except ValueError:
                log.error("invalid data in default layout")
                layout = {}
        return layout

    def get_layout_metadata(self, s, keyword):
        """get the TileManager layout
        """
        layout_dict = s.get(keyword, {})
        if not layout_dict:
            layout_dict = self.get_default_layout().get(keyword, {})
        return layout_dict

    def restore_layout_and_viewers(self, s):
        viewers = self.get_layout_metadata(s, "viewers")
        if viewers:
            layout = self.get_layout_metadata(s, "layout")
            if layout:
                self.control.restore_layout(layout)
        else:
            viewers = [{'name':name, 'uuid':name} for name in self.default_viewers.split(",")]
        log.critical(viewers)

        viewer_metadata = {}
        for v in viewers:
            viewer_metadata[v['uuid']] = v
            log.debug("metadata: viewer[%s]=%s" % (v['uuid'], str(v)))

        self.create_viewers(viewer_metadata)

    def set_initial_focused_viewer(self):
        if self.focused_viewer is None:
            for viewer in self.viewers:
                if not self.control.in_sidebar(viewer.control):
                    break
            print(("setting focus to %s" % viewer))
            self.set_focused_viewer(viewer)
            self.force_focus(viewer)


    def from_metadata_dict(self, e):
        log.debug("metadata: %s" % str(e))
        viewers = e.get('viewers', [])
        log.debug("metadata: viewers=%s" % str(viewers))
        
        if not viewers:
            try:
                e_default = self.get_default_layout()
                print(("using defaults from template: template=%s" % str(e_default)))
            except OSError:
                log.error("No template for default layout; falling back to minimal setup.")
            else:
                e.update(e_default)
                viewers = e.get('viewers', [])
            log.debug("from layout: viewers=%s" % str(viewers))

        layout = e.get('layout', {})
        log.debug("metadata: layout=%s" % str(layout))

        viewer_metadata = {}
        for v in viewers:
            viewer_metadata[v['uuid']] = v
            log.debug("metadata: viewer[%s]=%s" % (v['uuid'], str(v)))

        log.debug("task arguments: '%s'" % self.task_arguments)
        if self.task_arguments or not viewer_metadata:
            names = self.task_arguments if self.task_arguments else self.default_viewers
            log.debug("overriding viewers: %s" % str(names))
            override_viewer_metadata = {}
            for viewer_name in names.split(","):
                if viewer_name == "emulator":
                    continue
                override_viewer_metadata[viewer_name.strip()] = {}
                log.debug("metadata: clearing viewer[%s] because specified in task args" % (viewer_name.strip()))
            if override_viewer_metadata:
                # found some specified viewers, so override the default layout
                viewer_metadata = override_viewer_metadata
                layout = {}  # empty layout so it isn't cluttered with unused windows

        linked_bases = {}
        for b in e.get('linked bases', []):
            base = LinkedBase(editor=self)
            base.from_metadata_dict(b)
            linked_bases[base.uuid] = base
            log.debug("metadata: linked_base[%s]=%s" % (base.uuid, base))
        uuid = e.get("center_base", None)
        try:
            self.center_base = linked_bases[uuid]
        except KeyError:
            self.center_base = LinkedBase(editor=self)
        self.create_viewers(layout, viewer_metadata, e, linked_bases)
        viewer = None
        if 'focused viewer' in e:
            u = e['focused viewer']
            viewer = self.find_viewer_by_uuid(u)
        if viewer is None:
            for viewer in self.viewers:
                if not self.control.in_sidebar(viewer.control):
                    break
        print(("setting focus to %s" % viewer))
        print("center base", self.center_base.segment_uuid)
        self.set_focused_viewer(viewer)
        self.task.segments_changed = self.document.segments

    def serialize_session(self, s):
        s["layout"] = self.control.calc_layout()
        s["viewers"] = []
        bases = {}
        for v in self.viewers:
            b = v.linked_base
            bases[b.uuid] = b
            e = {"linked base": v.linked_base.uuid}
            v.serialize_session(e)
            s["viewers"].append(e)
        if self.center_base is not None:
            bases[self.center_base.uuid] = self.center_base
            s["center_base"] = self.center_base.uuid
        else:
            s["center_base"] = None
        s["linked bases"] = []
        for u, b in bases.items():
            e = {}
            b.serialize_session(e)
            s["linked bases"].append(e)
        s["focused viewer"] = self.focused_viewer.uuid

    #### viewer utilities

    def reconfigure_viewers(self):
        self.update_viewer_names()

    def update_viewer_names(self):
        for viewer in self.viewers:
            viewer.update_caption()
        self.control.update_captions()

    def add_viewer(self, viewer_or_viewer_cls, linked_base=None, replace_uuid=None):
        if hasattr(viewer_or_viewer_cls, "control"):
            viewer = viewer_or_viewer_cls
        else:
            if linked_base is None:
                if self.focused_viewer is not None:
                    linked_base = self.focused_viewer.linked_base
                else:
                    raise RuntimeError("Creating a viewer with no linked base and no focused viewer")
            viewer = viewer_or_viewer_cls.viewer_factory(self.control, linked_base)
        print("VIWER", viewer, type(viewer_or_viewer_cls), linked_base)
        print("SEGMENT", linked_base.segment)
        if replace_uuid:
            viewer.uuid = replace_uuid
        self.viewers.append(viewer)
        self.control.add(viewer.control, viewer.uuid)
        viewer.recalc_data_model()
        self.update_pane_names()
        return viewer

    def add_viewer_by_name(self, viewer_name):
        viewer_cls = find_viewer_class_by_name(viewer_name)
        self.add_viewer(viewer_cls)

    def replace_viewer(self, viewer_to_replace, new_viewer, linked_base):
        # self.viewers.remove(viewer_to_replace)
        # viewer_to_replace.prepare_for_destroy()
        created_viewer = self.add_viewer(new_viewer, linked_base, replace_uuid=viewer_to_replace.uuid)
        # self.set_focused_viewer(created_viewer)
        # del viewer_to_replace

    def replace_focused_viewer(self, linked_base):
        viewer_cls = self.focused_viewer.__class__
        self.replace_viewer(self.focused_viewer, viewer_cls, linked_base)

    def verify_center_base_is_used(self):
        if self.center_base is not None:
            for v in self.viewers:
                if v.linked_base == self.center_base:
                    break
            else:
                self.center_base = None
        print(f"Center base is: {self.center_base}")

    def unlink_viewer(self):
        number = self.focused_viewer.linked_base.segment_uuid
        base = LinkedBase(editor=self)
        base.view_segment_uuid(number)
        self.replace_focused_viewer(base)
        self.verify_center_base_is_used()

    def link_viewer(self):
        if self.center_base is None:
            self.center_base = self.focused_viewer.linked_base
        else:
            base = self.center_base
            self.replace_focused_viewer(base)

    def check_viewer_center_base(self, viewer_control, toggle_id):
        try:
            v = viewer_control.segment_viewer
        except AttributeError:
            state = True
        else:
            state = v.linked_base == self.center_base
        return state

    def create_viewers(self, viewer_metadata):
        # Create a set of viewers from a list
        import pprint
        log.debug("viewer_metadata: %s" % str(list(viewer_metadata.keys())))

        #self.document.find_initial_visible_segment(self.center_base)

        layer = 0
        viewers = list(viewer_metadata.keys())

        center_base_used = False
        while not self.viewers:
            for uuid in viewers:
                log.debug("loading viewer: %s" % uuid)
                e = viewer_metadata[uuid]
                if e:
                    viewer_type = e['name']
                else:  # either not a uuid or an unknown uuid
                    viewer_type = uuid  # try the value of 'uuid' as a viewer name
                try:
                    viewer_cls = find_viewer_class_by_name(viewer_type)
                except ValueError:
                    log.error("unknown viewer %s, uuid=%s" % (viewer_type, uuid))
                    continue
                log.debug("identified viewer: %s" % viewer_cls)

                if e:
                    try:
                        linked_base = self.linked_bases[e['linked base']]
                    except KeyError:
                        linked_base = self.center_base
                    log.debug("recreating viewer %s: %s" % (viewer_type, uuid))
                else:  # either not a uuid or an unknown uuid
                    linked_base = viewer_cls.calc_segment_specific_linked_base(self)
                    if linked_base is None:
                        linked_base = self.center_base
                    log.debug("using default metadata for %s" % (viewer_type))
                if linked_base == self.center_base:
                    center_base_used = True

                log.debug("creating viewer %s (%s) with linked base %s" % (uuid, viewer_type, str(linked_base)))
                viewer = viewer_cls.viewer_factory(self.control, linked_base, uuid, e)
                log.debug("created viewer %s (%s)" % (viewer.uuid, viewer.name))

                self.viewers.append(viewer)
                if not self.control.replace_by_uuid(viewer.control, viewer.uuid):
                    log.debug("viewer %s not found, adding in new pane" % viewer.uuid)
                    self.control.add(viewer.control, viewer.uuid)

            if not self.viewers:
                # just load default hex editor if nothing has been created
                viewers = ['hex']
                first = False

        if not center_base_used:
            self.center_base = None
        self.update_pane_names()

    def find_viewer_by_uuid(self, u):
        for v in self.viewers:
            if u == v.uuid:
                return v
        return None

    #### wx event handlers

    def force_focus(self, viewer):
        self.control.force_focus(viewer.uuid)
        c = viewer.control
        if not c.caret_handler.has_carets:
            c.caret_handler.move_current_caret_to_index(c.table, 0)
            flags = c.create_mouse_event_flags()
            flags.carets_to_indexes = [(0, -1, -1)]
            flags.source_control = c
            self.linked_base.sync_caret_to_index_event(flags=flags)
            wx.CallAfter(c.SetFocus)
        self.update_pane_names()
        viewer.update_toolbar()

    def set_focused_viewer(self, viewer):
        self.focused_viewer = viewer
        self.focused_viewer_changed_event(viewer)
        self.caret_handler = viewer.linked_base
        viewer.linked_base.segment_selected_event(viewer.linked_base.segment_uuid)
        self.set_menu_for_viewer(viewer)

    def set_menu_for_viewer(self, viewer):
        """Update menubar/toolbar for the current viewer.

        Viewers can have menu items that only apply when that viewer is
        currently focused, so this will change the UI every time a new viewer
        is selected.

        As currently implemented, all possible menu/tool items are listed in
        the class attributes `menubar_desc` and `toolbar_desc`, and the viewer
        prunes the list to remove items that aren't applicable. However, the
        entire description could be built on the fly by rewriting the call to
        `SegmentViewer.prune_menubar` to build a new description list from
        scratch.
        """
        # Override the class attributes with instance attributes
        self.menubar_desc = viewer.prune_menubar(self.__class__.menubar_desc)
        self.toolbar_desc = viewer.prune_toolbar(self.__class__.toolbar_desc)

        # force rebuild of UI
        self.frame.make_active(self, True)

    def on_viewer_active(self, evt):
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            v = evt.child.segment_viewer
            if v == self.focused_viewer:
                log.debug("on_pane_active: already current viewer %s" % v)
            else:
                log.debug("on_pane_active: activated viewer %s %s" % (v, v.window_title))
                self.set_focused_viewer(v)

    def on_viewer_close(self, evt):
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            log.debug("on_pane_close: closed viewer %s %s" % (v, v.window_title))

            # Keep a reference to the linked base
            linked_base_save = v.linked_base

            self.viewers.remove(v)
            v.prepare_for_destroy()

            from .. import viewers as omnivore8bit_viewers
            if not self.viewers:
                v = self.add_viewer(omnivore8bit_viewers.PlaceholderViewer, linked_base_save)
            self.set_focused_viewer(self.viewers[0])
            del v

    def on_viewer_replace(self, evt):
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            log.debug("on_viewer_replace: closing viewer %s %s for replacement" % (v, v.window_title))

            # Keep a reference to the linked base
            linked_base_save = v.linked_base

            self.viewers.remove(v)
            v.prepare_for_destroy()
            self.set_focused_viewer(evt.replacement_child.segment_viewer)
            del v

    def on_viewer_link(self, evt):
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            toggle = evt.GetInt()
            desired_state = evt.IsChecked()
            log.debug("on_viewer_replace: linking viewer %s %s: %s" % (v, v.window_title, desired_state))
            if desired_state:
                print("LINKED!")
                self.link_viewer()
            else:
                print("UNLINKED!")
                self.unlink_viewer()



""" Action definitions for Jumpman Task

"""
import os
import sys

import wx
import numpy as np

from sawx.ui.dialogs import prompt_for_hex, prompt_for_string, ChooseOnePlusCustomDialog
from sawx.utils.textutil import text_to_int

from .. import errors
from ..viewers.commands import SetValueCommand
from ..viewers.actions import ViewerAction

from .parser import DrawObjectBounds, is_valid_level_segment

import logging
progress_log = logging.getLogger("progress")

import logging
log = logging.getLogger(__name__)


def trigger_dialog(event, segment_viewer, obj):
    model = segment_viewer.linked_base.jumpman_playfield_model
    possible_labels = model.get_triggers()
    label = model.get_trigger_label(obj.trigger_function)
    if label is None and obj.trigger_function:
        custom_value = "%04x" % obj.trigger_function
    else:
        custom_value = ""
    dlg = ChooseOnePlusCustomDialog(event.task.window.control, list(possible_labels.keys()), label, custom_value, "Choose Trigger Function", "Select one trigger function or enter custom address", "Trigger Addr (hex)")
    if dlg.ShowModal() == wx.ID_OK:
        label, addr = dlg.get_selected()
        if label is not None:
            addr = possible_labels[label]
        else:
            try:
                addr = text_to_int(addr, "hex")
            except ValueError:
                event.task.window.error("Invalid address %s" % addr)
                addr = None
    else:
        addr = None
    dlg.Destroy()
    return addr

class JumpmanBaseCommand(SetValueCommand):
    def set_undo_flags(self, flags):
        # FIXME: need to add new flags to control rebuilding of objects, etc?
        flags.byte_values_changed = True


class CreateObjectCommand(JumpmanBaseCommand):
    short_name = "create_jumpman_obj"
    ui_name = "Create Object"


class MoveObjectCommand(JumpmanBaseCommand):
    short_name = "move_jumpman_obj"
    ui_name = "Move Object"


class FlipVerticalCommand(JumpmanBaseCommand):
    short_name = "vflip_jumpman_obj"
    ui_name = "Flip Vertically"


class FlipHorizontalCommand(JumpmanBaseCommand):
    short_name = "hflip_jumpman_obj"
    ui_name = "Flip Horizontally"


class ClearTriggerCommand(JumpmanBaseCommand):
    short_name = "cleartrigger_jumpman_obj"
    ui_name = "Clear Trigger Function"


class SetTriggerCommand(JumpmanBaseCommand):
    short_name = "settrigger_jumpman_obj"
    ui_name = "Set Trigger Function"


class AssemblyChangedCommand(JumpmanBaseCommand):
    short_name = "jumpman_custom_code"
    ui_name = "Reassemble Custom Code"



"""Segment actions
"""
import os
import sys

import wx

from sawx.ui.dialogs import prompt_for_string

from . import ViewerAction

from .. import commands
from ... import errors

import logging
log = logging.getLogger(__name__)


def prompt_for_comment(e, s, ranges, desc):
    existing = s.get_first_comment(ranges)
    text = prompt_for_string(e.control, desc, "Add Comment", existing)
    if text is not None:
        cmd = commands.SetCommentCommand(s, ranges, text)
        e.process_command(cmd)


class comment_add(ViewerAction):
    """Add a text comment to a byte location.

    A comment is associated with a single byte, so although a range can be
    selected, the comment is applied to only the first byte in the range.

    Bytes with comments will be highlighted in all displays.
    """
    name = 'Add Comment'

    def calc_enabled(self, action_key):
        return True

    def is_range(self):
        return self.viewer.control.caret_handler.has_selection

    def get_index(self):
        c = self.viewer.control.caret_handler.current
        index, _ = self.viewer.control.table.get_index_range(c.rc[0], c.rc[1])
        return index

    def perform(self, action_key):
        e = self.editor
        s = e.segment
        if self.is_range():
            ranges = s.get_style_ranges(selected=True)
            if len(ranges) == 1:
                desc = "Enter comment for first byte of range:\n%s" % self.viewer.get_label_of_first_byte(ranges)
            else:
                desc = "Enter comment for first byte of each range:\n%s" % self.viewer.get_label_of_first_byte(ranges)
        else:
            ranges = []
        if not ranges:
            index = self.get_index()
            ranges = [(index, index+1)]
            desc = "Enter comment for location %s" % index
        prompt_for_comment(e, s, ranges, desc)


class comment_remove(ViewerAction):
    """Remove any comments that are in the selected range, or if no selection
    from the current caret position.

    """
    name = 'Remove Comment'
    accelerator = 'Shift+Alt+C'

    def perform(self, event):
        e = self.editor
        s = e.segment
        if e.can_copy:
            ranges = s.get_style_ranges(selected=True)
        else:
            index = e.caret_index
            ranges = [(index, index+1)]
        if ranges:
            cmd = commands.ClearCommentCommand(s, ranges)
            e.process_command(cmd)



""" Text editor sample task

"""
# Enthought library imports.
from pyface.action.api import Separator, Group
from pyface.tasks.api import Task, TaskWindow, TaskLayout, PaneItem, IEditor, \
    IEditorAreaPane, EditorAreaPane, Editor, DockPane, HSplitter, VSplitter
from pyface.tasks.action.api import SMenuBar, SMenu, SToolBar, SchemaAddition
from traits.api import on_trait_change, Property, Instance, Any, Event, Int, Bool

from .. import known_emulators

from sawx.framework.task import FrameworkTask
from sawx.framework import actions as fa
from sawx.framework.toolbar import get_toolbar_group
from . import actions as ea
from .preferences import EmulatorPreferences
from .editor import EmulatorEditor
from ..byte_edit.task import ByteEditTask
from ..byte_edit import actions as ba
from ..viewers import actions as va
from ..arch import fonts, colors, machine
from ..utils.segmentutil import iter_known_segment_parsers


class EmulatorTask(ByteEditTask):
    """The emulator task supports a variety of 8-bit system emulators, 
    providing a common front-end and debugger to all of them.

    All data viewers from the `ByteEditTask`__ are available here, as well as
    viewers that only make sense when examining emulator internals and output.

    Currently there are emulators for systems:

    * Atari 8-bit computer

      * Atari 400/800
      * Atari XL

    * Generic 6502

    * Apple ][+ (very limited support)
    """

    new_file_text = []

    editor_id = "omnivore.emulator"

    pane_layout_version = ""

    #### Task interface #######################################################

    id = editor_id

    name = 'Emulator'

    preferences_helper = EmulatorPreferences

    #### Menu events ##########################################################

    ui_layout_overrides = {
        "menu": {
            "order": ["File", "Edit", "View", "Bytes", "Segment", "Emulation", "Documents", "Window", "Help"],
            "View": ["PredefinedGroup", "ProcessorGroup", "AssemblerGroup", "MemoryMapGroup", "ColorGroup", "FontGroup", "BitmapGroup", "SizeGroup", "ChangeGroup", "ConfigGroup", "ToggleGroup", "TaskGroup", "DebugGroup"],
            "Bytes": ["HexModifyGroup"],
            "Segment": ["ListGroup", "ActionGroup", "LabelGroup"],
            "Disk Image": ["ParserGroup", "ActionGroup"],
            "Emulation":  ["ConfigGroup", "CommandGroup"],
        },
    }


    #### FrameworkTask interface

    def get_editor(self, task_arguments="", **kwargs):
        """ Opens a new empty window
        """
        editor = EmulatorEditor(task_arguments=task_arguments)
        return editor

    #### Menu and toolbar definition

    def get_actions_Menu_Emulation_CommandGroup(self):
        return [
            ea.PauseResumeAction(),
            ea.PreviousSaveStateAction(),
            ea.NextSaveStateAction(),
            Separator(),
            ea.StepAction(),
            ea.StepIntoAction(),
            ea.StepOverAction(),
            ea.EndOfFrameAction(),
            ea.BreakVBIStart(),
            Separator(),
            ea.StartAction(),
            ea.SelectAction(),
            ea.OptionAction(),
            Separator(),
            ea.WarmstartAction(),
            ea.ColdstartAction(),
            ]

    #### file recognizer

    @classmethod
    def can_edit(cls, document):
        return hasattr(document, "emulator_type")

    @classmethod
    def get_match_score(cls, document):
        """Return a number based on how good of a match this task is to the
        incoming Document.
        
        0 = generic match
        ...
        10 = absolute match
        """
        return 10



import ctypes
import time

import numpy as np
np.set_printoptions(formatter={'int':hex})

from ..generic6502 import lib6502, Generic6502
from ..generic6502 import dtypes as d

import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)


debug_frames = False


class Crabapple(Generic6502):
    cpu = "6502"
    name = "crabapple"
    ui_name = "Crabapple 2+"

    mime_types = set(["application/vnd.apple2.bin"])

    def configure_emulator_defaults(self):
        lib6502.set_a2_emulation_mode(1)

    def generate_save_state_memory_blocks(self):
        cpu_offset = self.state_start_offset
        memory_offset = cpu_offset + d.CPU_DTYPE.itemsize
        segments = [
            (cpu_offset, d.CPU_DTYPE.itemsize, 0, "CPU Status"),
            (memory_offset, d.MAIN_MEMORY_SIZE, 0, "Main Memory"),
            (memory_offset + 0x400, 0x400, 0x400, "Text/Lo-res Page 1"),
            (memory_offset + 0x800, 0x400, 0x800, "Text/Lo-res Page 2"),
            (memory_offset + 0x2000, 0x2000, 0x2000, "Hi-res Page 1"),
            (memory_offset + 0x4000, 0x2000, 0x4000, "Hi-res Page 2"),
        ]
        self.save_state_memory_blocks.extend(segments)



import os
import sys

import numpy as np

import wx

from ..udis_fast import TraceInfo, flag_origin

from sawx.framework.enthought_api import EditorAction
from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg
import sawx.framework.clipboard as clipboard

from ..ui.segment_grid import SegmentGridControl, SegmentTable, SegmentGridTextCtrl
from .hex2 import HexEditControl
from ..arch.disasm import iter_disasm_styles
from ..utils import searchutil
from ..commands import SetIndexedDataCommand
from .actions import ViewerAction

from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)



class UdisFastTable(cg.HexTable):
    column_labels = ["Bytes", "Disassembly", "Comment"]
    column_sizes = [11, 18, 30]

    def __init__(self, linked_base):
        s = linked_base.segment
        cg.HexTable.__init__(self, s.data, s.style, 2, s.origin)
        self.lines = None
        self.num_rows = 0
        self.index_to_row = []
        self.end_addr = 0
        self.chunk_size = 256
        self.set_display_format(linked_base.cached_preferences)

    def set_display_format(self, prefs):
        for i, w in enumerate(prefs.disassembly_column_widths):
            if w > 0:
                self.__class__.column_pixel_sizes[i] = w
        self.set_fmt_hex(prefs.hex_format_character)

    def set_fmt_hex(self, fmt_char):
        self.fmt_hex1 = "%" + fmt_char
        self.fmt_hex2 = "%02" + fmt_char
        self.fmt_hex4 = "%04" + fmt_char

    def update_disassembly(self, segment, disassembly, index=0, refresh=False):
        self.segment = segment
        self.data = segment.data
        self.style = segment.style
        self.last_valid_index = len(self.data)
        self.disassembly = disassembly
        # cache some values for fewer deep references
        self.index_to_row = self.disassembly.info.index_to_row
        self.lines = self.disassembly.info
        self.jump_targets = self.disassembly.info.labels
        self.origin = self.disassembly.origin
        self.end_addr = self.disassembly.end_addr
        self.num_rows = self.lines.num_instructions
        # print(self.disassembly, self.num_rows)

    def get_index_range(self, r, c):
        try:
            try:
                line = self.lines[r]
            except IndexError:
                line = self.lines[-1]
            except TypeError:
                return 0, 0
            index = line.pc - self.origin
            return index, index + line.num_bytes
        except IndexError:
            if r >= self.num_rows:
                index = self.last_valid_index - 1
            else:
                index = 0
            return index, index

    def is_index_valid(self, index):
        return self.num_rows > 0 and index >= 0 and index <= self.last_valid_index

    def get_index_of_row(self, row):
        line = self.lines[row]
        index = line.pc - self.origin
        return index

    def get_start_end_index_of_row(self, row):
        line = self.lines[row]
        index = line.pc - self.origin
        return index, index + line.num_bytes

    def is_pc_valid(self, pc):
        index = pc - self.origin
        return self.is_index_valid(index)

    def index_to_row_col(self, index, col=1):
        try:
            row = self.index_to_row[index]
        except:
            try:
                row = self.index_to_row[-1]
            except IndexError:
                return 0, 0
        return row, col

    def get_next_caret_pos(self, row, col):
        col += 1
        if col >= self._cols:
            if row < self.num_rows - 1:
                row += 1
                col = 1
            else:
                col = self._cols - 1
        return (row, col)

    def get_next_editable_pos(self, row, col):
        if col < 1:
            col = 1
        elif col == 1:
            col = 1
            row += 1
        elif col == 2:
            col = 2
            row += 1
        return (row, col)

    def get_prev_caret_pos(self, row, col):
        col -= 1
        if col < 1:
            if row > 0:
                row -= 1
                col = self._cols - 1
            else:
                col = 1
        return (row, col)

    def get_page_index(self, index, segment_page_size, dir, grid):
        r, c = self.get_row_col(index)
        vr = grid.get_num_visible_rows() - 1
        r += (dir * vr)
        if r < 0:
            r = 0
        index, _ = self.get_index_range(r, 0)
        return index

    def get_pc(self, row):
        try:
            row = self.lines[row]
            return row.pc
        except IndexError:
            return 0
        except TypeError:
            return 0

    def get_value_style(self, row, col, operand_labels_start_pc=-1, operand_labels_end_pc=-1, extra_labels={}, offset_operand_labels={}, line=None):
        if line is None:
            line = self.lines[row]
        index = line.pc - self.origin
        style = 0
        for i in range(line.num_bytes):
            style |= self.style[index + i]
        text = self.calc_display_text(row, col, line, index)
        return text, style

    def calc_display_text(self, row, col, line=None, index=None):
        if line is None:
            line = self.lines[row]
            index = line.pc - self.origin
        if col == 0:
            if self.lines[row].flag == flag_origin:
                text = ""
            else:
                text = self.disassembly.format_data_list_bytes(index, line.num_bytes)
        else:
            text = self.disassembly.format_instruction(index, line)
            comment = self.disassembly.format_comment(index, line)
            if comment:
                text += " ; " + comment
        return text

    def get_label_at_index(self, index):
        row = self.index_to_row[index]
        return self.get_label_at_row(row)

    def get_label_at_row(self, row):
        addr = self.get_pc(row)
        return self.fmt_hex4 % addr


class DisassemblyGridControl(SegmentGridControl):
    def calc_default_table(self, linked_base):
        return UdisFastTable(linked_base)

    def calc_line_renderer(self):
        return cg.VirtualTableLineRenderer(self, 2, widths=[5,25], col_labels=['^Opcodes','^      Operand'])

    def recalc_view(self):
        v = self.segment_viewer
        v.restart_disassembly()
        cg.CompactGrid.recalc_view(self)
        if v.is_tracing:
            v.update_trace_in_segment()

    def get_disassembled_text(self, start=0, end=-1):
        return self.table.disassembly.get_disassembled_text(start, end)

    def encode_data(self, segment, linked_base):
        """Segment saver interface: take a segment and produce a byte
        representation to save to disk.
        """
        lines = self.table.disassembly.get_disassembled_text()
        text = os.linesep.join(lines) + os.linesep
        data = text.encode("utf-8")
        return data

    def extra_popup_actions(self, popup_data):
        actions = []
        addr_dest = self.table.disassembly.get_addr_dest(popup_data['row'], popup_data['col'])
        actions.extend(self.segment_viewer.linked_base.get_goto_actions_other_segments(addr_dest))
        actions.extend(self.segment_viewer.linked_base.get_goto_actions_same_byte(popup_data['index']))
        return actions

    ##### editing

    def advance_caret_position(self):
        self.handle_char_move_down(None, None)

    def verify_keycode_can_start_edit(self, c):
        if c == ord(";"):
            self.editing_type = "comment"
        else:
            self.editing_type = "code"
        return True

    def use_first_char_when_starting_edit(self):
        return self.editing_type != "comment"

    def process_edit(self, val):
        if self.editing_type == "comment":
            self.process_comment(val)
        else:
            try:
                self.process_mnemonic(val)
            except RuntimeError as e:
                log.error(f"Invalid assembly: {e}")
                self.segment_viewer.editor.window.error(str(e), "Invalid Assembly")

    def process_comment(self, val):
        ranges = self.get_selected_ranges_including_carets(self.caret_handler)
        cmd = SetCommentCommand(self.segment_viewer.segment, ranges, val)
        self.segment_viewer.editor.process_command(cmd)

    def process_mnemonic(self, val):
        t = self.table
        d = t.disassembly
        op = val.upper()
        ranges = self.get_selected_ranges_including_carets(self.caret_handler)
        max_possible = ranges[-1][1] + 16  # assuming max opcode length < 16 bytes
        data = np.zeros([max_possible], dtype=np.int16) - 1  # negative == unused

        # Have to calculate each line individually because PC relative
        # operations will have different opcode bytes depending on what the PC
        # is at the current location.
        for r in ranges:
            print(("processing range %s" % str(r)))
            row = t.index_to_row[r[0]]
            data_index = t.get_index_of_row(row)
            pc = t.get_pc(row)
            subset = None
            while data_index < r[1]:
                opcodes = d.assemble_text(pc, op)
                next_data_index = data_index + len(opcodes)
                print(("pc=%x op=%s data[%d:%d]=%s" % (pc, op, data_index, next_data_index, opcodes)))
                data[data_index:next_data_index] = opcodes
                data_index = next_data_index
                pc += len(opcodes)

        indexes = np.where(data >= 0)[0]
        byte_data = np.empty([len(indexes)], dtype=np.uint8)
        byte_data[:] = data[indexes]
        print(indexes)
        print(byte_data)
        cmd = SetIndexedDataCommand(self.segment_viewer.segment, indexes, byte_data, advance=True)
        self.segment_viewer.editor.process_command(cmd)


class CopyDisassemblyAction(ViewerAction):
    """Copy the disassembly text of the current selection to the clipboard.

    """
    name = 'Copy Disassembly Text'
    enabled_name = 'can_copy'

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        ranges = s.get_style_ranges(selected=True)
        lines = []
        try:
            for start, end in ranges:
                lines.extend(self.viewer.control.table.disassembly.get_disassembled_text(start, end))
        except IndexError:
            e.window.error("Disassembly tried to jump to an address outside this segment.")
            return
        text = os.linesep.join(lines) + os.linesep
        data_obj = wx.TextDataObject()
        data_obj.SetText(text)
        clipboard.set_clipboard_object(data_obj)

    def _update_enabled(self, ui_state):
        self.enabled = self.active_editor.focused_viewer.has_cpu


class CopyDisassemblyCommentsAction(ViewerAction):
    """Copy the text of the comments only, using the disassembly for line
    breaks. Any blank lines that appear in the disassembly are included in the
    copy.

    """
    name = 'Copy Disassembly Comments'
    enabled_name = 'can_copy'

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        ranges = s.get_style_ranges(selected=True)
        lines = []
        for start, end in ranges:
            for _, _, _, comment, _ in self.viewer.control.table.disassembly.iter_row_text(start, end):
                lines.append(comment)
        text = os.linesep.join(lines) + os.linesep
        data_obj = wx.TextDataObject()
        data_obj.SetText(text)
        clipboard.set_clipboard_object(data_obj)

    def _update_enabled(self, ui_state):
        self.enabled = self.active_editor.focused_viewer.has_cpu


class PasteDisassemblyCommentsAction(ViewerAction):
    name = 'Paste Disassembly Comments'
    tooltip = 'Paste text as comment lines'
    enabled_name = 'can_paste'
    accelerator = 'Shift-F6'

    def perform(self, event):
        self.active_editor.paste(PasteDisassemblyCommentsCommand, disasm=self.viewer.control.table.disassembly)


class PasteDisassemblyCommentsCommand(SetCommentCommand):
    """Copy the text of the comments only, using the disassembly for line
    breaks. Any blank lines that appear in the disassembly are included in the
    copy.

    """
    name = 'Paste Disassembly Comments'

    def __init__(self, segment, serializer, disasm=None):
        ranges = serializer.dest_carets.selected_ranges
        if not ranges:
            # use the range from caret to end
            ranges = [(serializer.dest_carets.current.index, len(segment))]
        SetCommentCommand.__init__(self, segment, ranges, serializer.clipboard_data)
        self.comments = self.text.tobytes().splitlines()
        self.num_lines = len(self.comments)
        self.disasm = disasm

    def __str__(self):
        return "%s: %d line%s" % (self.ui_name, self.num_lines, "" if self.num_lines == 1 else "s")

    def clamp_ranges_and_indexes(self, editor):
        disasm = self.disasm
        count = self.num_lines
        comment_indexes = []
        clamped = []
        for start, end in self.ranges:
            index = start
            log.debug("starting range %d:%d" % (start, end))
            while index < end and count > 0:
                comment_indexes.append(index)
                pc = index + self.segment.origin
                log.debug("comment at %d, %04x" % (index, pc))
                try:
                    index = disasm.get_next_instruction_pc(pc)
                    count -= 1
                except IndexError:
                    count = 0
            clamped.append((start, index))
            if count <= 0:
                break
        return clamped, comment_indexes

    def change_comments(self, ranges, indexes):
        """Add comment lines as long as we don't go out of range (if specified)
        or until the end of the segment or the comment list is exhausted.

        Depends on a valid disassembly to find the lines; we are adding a
        comment for the first byte in each statement.
        """
        log.debug("ranges: %s" % str(ranges))
        log.debug("indexes: %s" % str(indexes))
        self.segment.set_comments_at_indexes(ranges, indexes, self.comments)
        self.segment.set_style_at_indexes(indexes, comment=True)



# Disassembly searcher uses the __call__ method to return the object because it
# needs extra info: the machine type & the disassembly list. Normal searchers
# just use the segment's raw data and returns itself in the constructor.
class DisassemblySearcher(searchutil.BaseSearcher):
    def __init__(self, viewer, panel):
        self.search_text = None
        self.matches = []
        self.panel = panel
        self.ui_name = viewer.machine.disassembler.name

    def __call__(self, editor, search_text):
        self.search_text = self.get_search_text(search_text)
        if len(self.search_text) > 0:
            self.matches = self.get_matches(editor)
            self.set_style(editor)
        else:
            self.matches = []
        return self

    def __str__(self):
        return "disasm matches: %s" % str(self.matches)

    def get_search_text(self, text):
        return text

    def get_matches(self, editor):
        matches = self.panel.search(self.search_text, editor.last_search_settings.get('match_case', False))
        return matches


class DisassemblyViewer(SegmentViewer):
    name = "disassembly"

    ui_name = "Disassembly"

    control_cls = DisassemblyGridControl

    has_cpu = True

    has_hex = True

    copy_special = [CopyDisassemblyAction, CopyDisassemblyCommentsAction]

    paste_special = [PasteDisassemblyCommentsAction]

    current_disassembly_ = Any(None)

    trace = Instance(TraceInfo)

    # trait defaults

    def _trace_default(self):
        return TraceInfo()

    # properties

    @property
    def window_title(self):
        return self.machine.disassembler.name + " (" + self.machine.memory_map.name + ")"

    @property
    def searchers(self):
        return [DisassemblySearcher(self, self.control)]

    # @on_trait_change('machine.disassembler_change_event')
    def do_disassembler_change(self, evt):
        log.debug("do_disassembler_change for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        if evt is not Undefined:
            self.clear_disassembly()
            self.restart_disassembly()
            self.linked_base.editor.update_pane_names()

    # @on_trait_change('linked_base.editor.document.byte_values_changed')
    def byte_values_changed(self, index_range):
        log.debug("byte_values_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.restart_disassembly(index_range)

    # @on_trait_change('linked_base.editor.document.byte_style_changed')
    def byte_style_changed(self, index_range):
        log.debug("byte_style_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.restart_disassembly(index_range)

    def recalc_data_model(self):
        self.clear_disassembly()
        self.restart_disassembly()

    def do_priority_level_refresh(self):
        index, row1, offset1, col1 = self.control.calc_primary_caret_visible_info()
        self.recalc_data_model()
        index2, row2, offset2, col2 = self.control.calc_primary_caret_visible_info()
        new_first_row = row2 - offset1
        log.debug(f"keep caret in same place: before={row1} {offset1}, after={row2} {offset2}, new={new_first_row}")
        self.control.move_viewport_origin((new_first_row, -1))
        self.refresh_view(True)
        self.frame_count = 0

    ##### UdisFast interface

    def create_disassembler(self):
        prefs = self.linked_base.cached_preferences
        d = self.machine.get_disassembler(prefs.hex_grid_lower_case, prefs.assembly_lower_case, self.document.document_memory_map, self.segment.memory_map)
        for i, name in iter_disasm_styles():
            d.add_chunk_processor(name, i)
        return d

    @property
    def current_disassembly(self):
        if self.current_disassembly_ is None:
            d = self.create_disassembler()
            self.current_disassembly_ = d
        return self.current_disassembly_

    def clear_disassembly(self):
        self.current_disassembly_ = None

    def restart_disassembly(self, index=None):
        self.current_disassembly.disassemble_segment(self.segment)
        self.control.table.update_disassembly(self.segment, self.current_disassembly)

    ##### disassembly tracing

    def start_trace(self):
        self.trace_info = TraceInfo()
        self.is_tracing = True
        self.update_trace_in_segment()
        self.linked_base.force_refresh()

    def get_trace(self, save=False):
        if save:
            kwargs = {'user': True}
        else:
            kwargs = {'match': True}
        s = self.segment
        mask = s.get_style_mask(**kwargs)
        style = s.get_style_bits(**kwargs)
        is_data = self.trace_info.marked_as_data
        size = min(len(is_data), len(s))
        trace = is_data[s.origin:s.origin + size] * style
        if save:
            # don't change data flags for stuff that's already marked as data
            s = self.segment
            already_data = np.logical_and(s.style[0:size] & user_bit_mask > 0, trace > 0)
            indexes = np.where(already_data)[0]
            previous = s.style[indexes]
            trace[indexes] = previous
        return trace, mask

    def update_trace_in_segment(self, save=False):
        trace, mask = self.get_trace(save)
        s = self.segment
        size = len(trace)
        s.style[0:size] &= mask
        s.style[0:size] |= trace

    def trace_disassembly(self, pc):
        self.current_disassembly.fast.trace_disassembly(self.trace_info, [pc])
        self.update_trace_in_segment()

    ##### popup commands

    def calc_viewer_popup_actions(self, popup_data):
        return []



import os
import time

import wx
import numpy as np
from atrip import SegmentData, DefaultSegment, DefaultSegmentParser, SegmentParser

# Enthought library imports.
from traits.api import Trait, Any, List, Event, Dict, Property, Bool, Int, String, Float, Undefined

from .. import find_emulator, guess_emulator, default_emulator, UnknownEmulatorError, EmulatorError

from ..document import DiskImageDocument

import logging
log = logging.getLogger(__name__)


class EmulatorSegmentParser(SegmentParser):
    menu_name = "Emulator Save State"
    def parse(self):
        r = self.segment_data
        self.segments.append(self.collection(r, 0, name=self.menu_name))
        for start, count, offset, name in self.save_state_memory_blocks:
            if count > 0:
                print(f"creating emulator segment {name} at {hex(start)}:{hex(start + count)}")
                self.segments.append(DefaultSegment(r[start:start + count], offset, name))

def segment_parser_factory(save_state_memory_blocks):
    cls = type('EmulatorSegmentParser', (EmulatorSegmentParser, SegmentParser), dict(save_state_memory_blocks = save_state_memory_blocks))
    return cls


class EmulationTimer(wx.Timer):
    def __init__(self, document):
        self.document = document
        wx.Timer.__init__(self)

    def Notify(self):
        self.document.ready_for_next_frame()


class EmulationDocument(DiskImageDocument):

    # Class attributes

    emulation_timer = None

    metadata_extension = ".omniemu"

    # Traits

    source_document = Any(None)

    boot_segment = Any(None)

    emulator_type = Any(Undefined)

    emulator = Any(None)

    skip_frames_on_boot = Int(0)

    # Update the graphic screen
    emulator_update_screen_event = Event

    framerate = Float(1/60.0)

    tickrate = Float(1/60.0)

    last_update_time = Float(0.0)

    ##### trait default values

    #### object methods

    def __str__(self):
        return f"EmulationDocument: id={self.document_id}, mime={self.metadata.mime}, {self.metadata.uri}. {self.emulator_type} source={self.source_document}"

    @classmethod
    def create_document(cls, source_document, emulator_type, skip_frames_on_boot=False, extra_metadata=None):
        try:
            emu_cls = find_emulator(emulator_type)
        except UnknownEmulatorError:
            if not emulator_type:
                # if no value specified, try to determine from binary data
                try:
                    emu_cls = guess_emulator(source_document)
                except UnknownEmulatorError:
                    emu_cls = default_emulator
            else:
                # if emulator name specified but not known, return error
                raise RuntimeError(f"Unknown emulator {emulator_type}")
        emu = emu_cls()
        emu.configure_emulator()
        log.debug(f"emulator changed to {emu}")
        doc = cls(emulator_type=emulator_type, emulator=emu, source_document=source_document)
        if extra_metadata:
            doc.restore_save_points_from_dict(extra_metadata)
        return doc

    @property
    def timer_delay(self):
        # wxpython delays are in milliseconds
        return self.tickrate * 1000

    @property
    def emulator_running(self):
        return self.emulation_timer is not None and self.emulation_timer.IsRunning()

    @property
    def emulator_paused(self):
        return self.emulation_timer is not None and not self.emulation_timer.IsRunning()

    #### serialization methods

    def restore_extra_from_dict(self, e):
        DiskImageDocument.restore_extra_from_dict(self, e)
        self.restore_save_points_from_dict(e)

    def restore_save_points_from_dict(self, e):
        # emulator type will already be set at document creation time
        if 'frame_history' in e:
            self.emulator.frame_history.restore_from_dict(e)
            self.emulator.configure_io_arrays()
            self.create_segments()
        if 'current_frame' in e:
            self.emulator.restore_from_dict(e['current_frame'])
        self.skip_frames_on_boot = e.get('skip_frames_on_boot', 0)

    def serialize_extra_to_dict(self, mdict):
        DiskImageDocument.serialize_extra_to_dict(self, mdict)

        mdict["emulator_type"] = self.emulator.name
        mdict.update(self.emulator.frame_history.serialize_to_dict())
        mdict["current_frame"] = self.emulator.serialize_to_dict()
        mdict["skip_frames_on_boot"] = self.skip_frames_on_boot

    def save_to_uri(self, uri, editor, saver=None, save_metadata=True):
        # save both the source document and its .omnivore metadata and...
        self.source_document.save_to_uri(uri, editor, saver, save_metadata)

        # the emulator metadata .omniemu from this emulator document
        if save_metadata:
            self.save_metadata_to_uri(uri, editor)

    ##### Initial viewer defaults

    def find_initial_visible_segment(self, linked_base, default=0):
        for segment in self.segments:
            log.debug(f"Looking for Main Memory segment: {segment}")
            if segment.name == "Main Memory":
                linked_base.find_segment(segment, refresh=False)
                break
        else:
            log.warning("Didn't find Main Memory segment in emulator")

    #####

    def boot(self, segment=None):
        emu = self.emulator
        if not emu.has_save_points:
            emu.configure_emulator([])
            if segment is None:
                segment = self.source_document.segment_parser.image.create_emulator_boot_segment()
            elif segment.origin == 0:
                segment = emu.find_default_boot_segment(self.source_document.segments)

            if segment is not None:
                boot_data = segment.data
                origin = segment.origin
            else:
                raise EmulatorError(f"Can't find bootable segment in {self.source_document}")
            emu.boot_from_segment(segment)
            for i in range(self.skip_frames_on_boot):
                emu.next_frame()
        self.create_segments()
        self.start_timer()

    def load(self, segment=None):
        if segment is None:
            segment = self.source_document.collection
        emu = self.emulator
        emu.begin_emulation([], segment)
        for i in range(self.skip_frames_on_boot):
            emu.next_frame()
        self.create_segments()
        self.start_timer()

    def create_segments(self):
        emu = self.emulator
        self.raw_bytes = emu.raw_array
        self.style = np.zeros([len(self.raw_bytes)], dtype=np.uint8)
        self.parse_segments([segment_parser_factory(emu.save_state_memory_blocks)])
        log.debug("Segments after boot: %s" % str(self.segments))
        self.create_timer()

    ##### Emulator commands

    def create_timer(self):
        cls = self.__class__
        if cls.emulation_timer is None:
            cls.emulation_timer = EmulationTimer(self)

    def start_timer(self, repeat=False, delay=None, forceupdate=True):
        if not self.emulation_timer.IsRunning():
            self.emulation_timer.StartOnce(self.framerate * 1000)

    def stop_timer(self, repeat=False, delay=None, forceupdate=True):
        self.emulation_timer.Stop()

    def ready_for_next_frame(self):
        now = time.time()
        breakpoint = self.emulator.next_frame()
        frame_number = self.emulator.status['frame_number']
        log.debug(f"showing frame {frame_number}, breakpoint_id={breakpoint}")
        if breakpoint is None:
            self.emulator_update_screen_event = True
            self.priority_level_refresh_event = True
            after = time.time()
            delta = after - now
            if delta > self.framerate:
                next_time = self.framerate * .8
            elif delta < self.framerate:
                next_time = self.framerate - delta
            log.debug(f"now={now} show={after-now} delta={delta} framerate={self.framerate} next_time={next_time}")
            if next_time <= 0.001:
                log.warning("need to drop frames!")
                next_time = .001
            self.emulation_timer.StartOnce(next_time * 1000)
        else:
            self.emulator_update_screen_event = True
            self.priority_level_refresh_event = 100
            print(f"generating breakpoint event: {breakpoint} at cycles={self.emulator.cycles_since_power_on}")
            self.emulator_breakpoint_event = breakpoint
            self.stop_timer()
        self.last_update_time = now

    def pause_emulator(self):
        print("pause")
        emu = self.emulator
        self.stop_timer()
        emu.cpu_history_show_next_instruction()
        self.emulator_update_screen_event = True
        self.priority_level_refresh_event = 100

    def restart_emulator(self):
        print("restart")
        self.start_timer()
        self.emulator_update_screen_event = True

    def debugger_step(self):
        print("stepping")
        self.emulator.step_into(1)
        self.start_timer()

    def debugger_break_vbi_start(self, count=1):
        print("stepping")
        self.emulator.break_vbi_start(count)
        self.start_timer()

    def debugger_count_frames(self, number=1):
        print(f"counting {number} frames")
        self.emulator.count_frames(number)
        self.start_timer()

    def history_previous(self):
        emu = self.emulator
        try:
            desired = emu.get_previous_history(emu.current_frame_number)
        except IndexError:
            log.warning("No previous frame")
        else:
            emu.restore_history(desired)
            frame_number = self.emulator.status['frame_number']
            log.debug(f"showing frame {frame_number}")
            self.emulator_update_screen_event = True
            self.priority_level_refresh_event = 100

    def history_next(self):
        emu = self.emulator
        try:
            desired = emu.get_next_history(emu.current_frame_number)
        except IndexError:
            log.warning(f"No next frame: current = {emu.current_frame_number}")
        else:
            emu.restore_history(desired)
            frame_number = self.emulator.status['frame_number']
            log.debug(f"showing frame {frame_number}")
            self.emulator_update_screen_event = True
            self.priority_level_refresh_event = 100



import os

import wx

import numpy as np
from atrip import Segment, Container, style_bits

from . import parser as ju
from . import commands as jc
from ..arch.colors import powerup_colors

import logging
log = logging.getLogger(__name__)
drawlog = logging.getLogger("refresh")


class JumpmanPlayfieldModel(object):
    def __init__(self, linked_base):
        self.linked_base = linked_base
        self.possible_jumpman_segment = ju.is_valid_level_segment(self.segment)
        self.items_per_row = 160
        self.antic_lines = 88
        self.playfield = self.get_playfield_segment()
        self.pick_buffer = np.zeros((self.items_per_row * self.antic_lines), dtype=np.int32)
        self.coin_harvest_diff = -1
        self.num_ladders = -1
        self.num_downropes = -1
        self.num_coins = -1
        self.can_select_objects = False
        self.can_erase_objects = False
        self.assembly_source = ""
        self.assembly_error = ""
        self.custom_code = None
        self.manual_recompile_needed = False
        self.old_trigger_mapping = {}
        self.trigger_root = None
        self.current_level = None
        self.cached_screen = None
        self.valid_level = False
        self.force_refresh = False
        self.screen_state = None
        self.trigger_state = None

    @property
    def segment(self):
        return self.linked_base.segment

    def init_level_builder(self, segment_viewer):
        log.debug("init_level_builder")
        self.segment_viewer = segment_viewer
        self.possible_jumpman_segment = ju.is_valid_level_segment(self.segment)
        self.level_builder = ju.JumpmanLevelBuilder(self.linked_base.document.user_segments)
        self.cached_screen = None
        self.valid_level = False
        self.force_refresh = True
        self.screen_state = None
        self.trigger_state = None
        self.generate_display_objects()
        self.level_colors = self.calc_level_colors()
        self.draw_playfield(True)
        if self.manual_recompile_needed:
            self.compile_assembly_source()

    @property
    def mouse_mode(self):
        return self.segment_viewer.control.mouse_mode

    def generate_display_objects(self, resync=False):
        try:
            source, level_addr, harvest_addr = self.get_level_addrs()
            index = level_addr - source.origin
            self.level_builder.parse_level_data(source, level_addr, harvest_addr)
            self.force_refresh = True
            self.valid_level = True
            if self.trigger_root is not None:
                self.trigger_root = self.level_builder.find_equivalent_coin(self.trigger_root)
            if resync:
                self.mouse_mode.resync_objects()
        except RuntimeError:
            self.valid_level = False

    def get_playfield(self):
        data = np.empty(self.items_per_row * self.antic_lines, dtype=np.uint8)
        return data

    def get_playfield_segment(self, playfield=None):
        if playfield is None:
            playfield = self.get_playfield()
        c = Container(playfield)
        return Segment(c, origin=0x7000)

    def clear_playfield(self, playfield=None):
        if playfield is None:
            playfield = self.playfield
        playfield[:] = 8  # background is the 9th ANTIC color register (counting from zero)
        playfield.style[:] = 0

    def calc_level_colors(self):
        if self.valid_level:
            colors = self.segment[0x2a:0x33].copy()
            # on some levels, the bombs are set to color 0 because they are
            # cycled to produce a glowing effect, but that's not visible here
            # so we force it to be bright white
            fg = colors[4:8]
            fg[fg == 0] = 15
        else:
            colors = powerup_colors()
        return list(colors)

    def get_level_addrs(self):
        if not self.possible_jumpman_segment:
            raise RuntimeError
        source = self.segment
        start = source.origin
        level_addr = source[0x37] + source[0x38]*256
        harvest_addr = source[0x4e] + source[0x4f]*256
        log.debug("level def table: %x, harvest table: %x" % (level_addr, harvest_addr))
        last = source.origin + len(source)
        if level_addr > start and harvest_addr > start and level_addr < last and harvest_addr < last:
            return source, level_addr, harvest_addr
        raise RuntimeError

    def set_trigger_root(self, root):
        if root is not None:
            root = self.level_builder.find_equivalent_coin(root)
        log.debug(f"setting trigger root to {root}")
        self.trigger_root = root
        self.force_refresh = True
        self.trigger_state = None

    @property
    def is_editing_trigger(self):
        return self.trigger_root is not None

    def get_screen_state(self):
        if self.trigger_state is not None:
            return self.trigger_state
        return self.screen_state

    def set_current_screen(self, screen=None):
        if screen is None:
            screen = self.playfield
        else:
            self.cached_screen = None
        self.data = screen
        self.style = screen.style

    def redraw_current(self, screen, overlay_objects=[]):
        self.clear_playfield(screen)
        self.pick_buffer[:] = -1
        self.level_builder.set_harvest_offset(self.mouse_mode.get_harvest_offset())
        main_state = self.level_builder.draw_objects(screen, None, self.segment, highlight=overlay_objects, pick_buffer=self.pick_buffer)
        drawlog.debug("draw objects: %s" % self.level_builder.objects)
        drawlog.debug("highlight objects: %s" % overlay_objects)
        if main_state.missing_object_codes:
            log.error("missing draw codes: %s" % (sorted(main_state.missing_object_codes)))
        if self.trigger_root is not None:
            self.level_builder.fade_screen(screen)
            root = [self.trigger_root]
            self.level_builder.draw_objects(screen, root, self.segment, highlight=root, pick_buffer=self.pick_buffer)
            # change highlight to comment color for selected trigger coin so
            # you don't get confused with any objects actually selected
            old_highlight = np.where(screen.style == style_bits.selected_bit_mask)
            screen.style[old_highlight] |= style_bits.comment_bit_mask
            screen.style[:] &= (0xff ^ (style_bits.match_bit_mask|style_bits.selected_bit_mask))

            # replace screen state so that the only pickable objects are
            # those in the triggered layer
            self.pick_buffer[:] = -1
            trigger_state = self.level_builder.draw_objects(screen, self.trigger_root.trigger_painting, self.segment, highlight=overlay_objects, pick_buffer=self.pick_buffer)
            active_state = trigger_state
        else:
            active_state = main_state
            trigger_state = None
        return main_state, trigger_state, active_state

    def compute_image(self, force=False):
        self.set_current_screen()
        if force:
            self.force_refresh = True
        if self.force_refresh or self.cached_screen is None:
            self.screen_state, self.trigger_state, _ = self.redraw_current(self.playfield)
            self.cached_screen = self.playfield[:].copy()
            self.force_refresh = False
            self.update_harvest_state()
        else:
            self.playfield[:] = self.cached_screen
        self.set_current_screen()
        return

    def bad_image(self):
        self.set_current_screen()
        self.playfield[:] = 0
        self.playfield.style[:] = 0
        self.force_refresh = True
        s = self.playfield.container.style.reshape((self.antic_lines, -1))
        s[::2,::2] = style_bits.comment_bit_mask
        s[1::2,1::2] = style_bits.comment_bit_mask
        self.set_current_screen()
        return

    def draw_playfield(self, force=False):
        if self.valid_level:
            override = self.mouse_mode.calc_playfield_override()
            if override is not None:
                drawlog.debug("draw_playfield: using override screen")
                self.set_current_screen(override)
                return
            drawlog.debug("draw_playfield: computing image")
            self.compute_image()
            self.mouse_mode.draw_extra_objects(self.level_builder, self.playfield, self.linked_base.segment)
        else:
            self.bad_image()
        # self.mouse_mode.draw_overlay(bitimage)  # FIXME!

    ##### Object edit

    def get_save_location(self):
        if self.trigger_root is not None:
            equiv = self.level_builder.find_equivalent_coin(self.trigger_root)
            parent = equiv.trigger_painting
        else:
            parent = None
        return parent

    def delete_objects(self, objects):
        save_location = self.get_save_location()
        self.level_builder.delete_objects(objects, save_location)
        self.save_changes()

    def save_objects(self, objects, command_cls=jc.CreateObjectCommand):
        save_location = self.get_save_location()
        self.level_builder.add_objects(objects, save_location)
        self.save_changes(command_cls)

    def save_changes(self, command_cls=jc.MoveObjectCommand):
        source, level_addr, old_harvest_addr = self.get_level_addrs()
        level_data, harvest_addr, ropeladder_data, num_coins = self.level_builder.create_level_definition(level_addr, source[0x46], source[0x47])
        index = level_addr - source.origin
        ranges = [(0x18,0x2a), (0x3e,0x3f), (0x4e,0x50), (index,index + len(level_data))]
        pdata = np.empty([1], dtype=np.uint8)
        if self.coin_harvest_diff < 0:
            pdata[0] = num_coins
        else:
            pdata[0] = max(0, num_coins - self.coin_harvest_diff)
        hdata = np.empty([2], dtype=np.uint8)
        hdata.view(dtype="<u2")[0] = harvest_addr
        data = np.hstack([ropeladder_data, pdata, hdata, level_data])
        cmd = command_cls(source, ranges, data)
        self.segment_viewer.editor.process_command(cmd)
        self.cached_screen = None
        log.debug("saved changes, new objects=%s" % self.level_builder.objects)

    ##### Utilities

    def update_harvest_state(self):
        if not self.current_level.valid_level:
            return
        harvest_state = self.current_level.level_builder.get_harvest_state()
        self.num_ladders = len(harvest_state.ladder_positions)
        self.num_downropes = len(harvest_state.downrope_positions)
        self.num_coins = len(harvest_state.coins)

    def set_current_draw_pattern(self, pattern, control):
        try:
            iter(pattern)
        except TypeError:
            self.draw_pattern = [pattern]
        else:
            self.draw_pattern = pattern

    ##### Custom code handling

    def set_assembly_source(self, src, do_compile=True):
        """Assembly source file is required to be in the same directory as the
        jumpman disk image. It's also assumed to be on the local filesystem
        since pyatasm can't handle the virtual filesystem.
        """
        self.assembly_source = src
        if do_compile:
            self.manual_recompile_needed = False
            self.compile_assembly_source()
        else:
            self.manual_recompile_needed = True

    def compile_assembly_source(self):
        self.custom_code = None
        if not self.assembly_source:
            return
        self.linked_base.editor.metadata_dirty = True
        d = self.linked_base.document
        path = d.filesystem_path()
        if not path:
            self.assembly_error = f"Assembly error:\nPlease save the level before\ncompiling the assembly source"
            log.error(self.assembly_error)
            return
        dirname = os.path.dirname(d.filesystem_path())
        if dirname:
            filename = os.path.join(dirname, self.assembly_source)
            try:
                log.debug("compiling jumpman level code in %s" % filename)
                self.custom_code = ju.JumpmanCustomCode(filename)
                self.manual_recompile_needed = False
            except SyntaxError as e:
                self.assembly_error = f"Assembly error:\n{str(e)}"
                log.error(self.assembly_error)
                self.manual_recompile_needed = True
            except ImportError:
                self.assembly_source = ""
                self.assembly_error = f"Assembly error:\nPlease install pyatasm to\ncompile custom code"
                log.error(self.assembly_error)
                self.old_trigger_mapping = dict()
            else:
                self.assembly_error = ""
            if self.custom_code:
                self.update_trigger_mapping()
                self.save_assembly()

    @property
    def custom_code_info(self):
        try:
            if self.assembly_error:
                return self.assembly_error
            return self.custom_code.info
        except AttributeError:
            return "No custom code"

    @property
    def action_vector_info(self):
        try:
            if self.assembly_error:
                return ""
            return self.custom_code.vector_summary
        except AttributeError:
            return "No custom code"

    @property
    def coin_trigger_info(self):
        try:
            if self.assembly_error:
                return ""
            return self.custom_code.coin_trigger_summary
        except AttributeError:
            return "No custom code"

    @property
    def other_label_info(self):
        try:
            if self.assembly_error:
                return ""
            return self.custom_code.label_summary
        except AttributeError:
            return "No custom code"

    def save_assembly(self):
        log.debug("save_assembly: code=%s" % self.custom_code)
        code = self.custom_code
        if not code:
            return
        source, level_addr, harvest_addr = self.get_level_addrs()
        ranges, data = code.get_ranges(source)
        cmd = jc.MoveObjectCommand(source, ranges, data)
        self.linked_base.editor.process_command(cmd)

    def update_trigger_mapping(self):
        # only create old trigger mapping if one doesn't exist
        if not self.old_trigger_mapping:
            self.old_trigger_mapping = dict(self.get_triggers())
        else:
            old_map = self.old_trigger_mapping
            new_map = self.get_triggers()
            if old_map != new_map:
                log.debug("UPDATING trigger map!")
                log.debug("old map %s" % old_map)
                log.debug("new_map %s" % new_map)
                self.current_level.update_triggers(old_map, new_map)
                # FIXME: what about undo and the trigger mapping?
                self.current_level.save_changes(AssemblyChangedCommand)
                self.old_trigger_mapping = new_map

    def update_harvest_state(self):
        if not self.valid_level:
            return
        harvest_state = self.level_builder.get_harvest_state()
        self.num_ladders = len(harvest_state.ladder_positions)
        self.num_downropes = len(harvest_state.downrope_positions)
        self.num_coins = len(harvest_state.coins)

    ##### Jumpman coin triggers

    def get_triggers(self):
        if self.custom_code is None and self.manual_recompile_needed == False:
            self.compile_assembly_source()
        code = self.custom_code
        if code is None:
            return {}
        return code.triggers

    def get_trigger_label(self, addr):
        rev_old_map = {v: k for k, v in self.old_trigger_mapping.items()}
        if addr in rev_old_map:
            return rev_old_map[addr]
        return None


import os
import sys

import numpy as np

import wx

from atrip import DefaultSegment

from ..disassembler import flags
from ..disassembler import dtypes as dd

from sawx.ui import compactgrid as cg
from ..editors.linked_base import VirtualTableLinkedBase

from ..ui.segment_grid import SegmentGridControl

from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)



class InstructionHistoryTable(cg.VirtualTable):
    column_labels = ["^Instruction", "^Result"]
    column_sizes = [21, 12]

    def __init__(self, linked_base):
        self.virtual_linked_base = linked_base
        s = linked_base.segment
        self.current_num_rows = 0
        self.history_entries = None
        self.visible_history_start_row = 0
        self.visible_history_lookup_table = None
        cg.VirtualTable.__init__(self, len(self.column_labels), s.origin)

    def calc_num_rows(self):
        return self.current_num_rows

    def calc_last_valid_index(self):
        return self.current_num_rows * self.items_per_row

    def get_label_at_index(self, index):
        row = (index // self.items_per_row) - self.visible_history_start_row
        if row < 0:
            return "----"
        try:
            emu = self.virtual_linked_base.emulator
            return "%04x" % (emu.cpu_history[row][0])
        except IndexError:
            return "----"

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        emu = self.virtual_linked_base.emulator
        for line in range(start_line, last_line, step):
            h = emu.cpu_history[line]
            t = h['disassembler_type']
            if t == flags.DISASM_NEXT_INSTRUCTION:
                h = h.view(dtype=dd.HISTORY_BREAKPOINT_DTYPE)
                t = h['disassembler_type_cpu']
            if t == flags.DISASM_ATARI800_HISTORY:
                h = h.view(dtype=dd.HISTORY_ATARI800_DTYPE)
                x = h['antic_xpos']
                y = h['antic_ypos']
                if x & 0x80:
                    x = x & 0x7f
                    y = y | 0x100
                yield "%3d %3d" % (y, x)
            elif t == flags.DISASM_FRAME_START or t == flags.DISASM_FRAME_END:
                h = h.view(dtype=dd.HISTORY_FRAME_DTYPE)
                f = h['frame_number']
                yield "f%d" % (f)
            else:
                yield "%d" % (emu.cpu_history[line][0])

    def calc_row_label_width(self, view_params):
        return view_params.calc_text_width("256 256")

    def get_value_style(self, row, col):
        t = self.parsed
        if t is None:
            return "", 0
        try:
            text = t[row - self.visible_history_start_row][col]
        except IndexError:
            print(f"tried row {row} out of {self.visible_history_lookup_table}")
            text = f"row {row} out of bounds"
        # style = s.style[index]
        style = 0
        return text, style

    def prepare_for_drawing(self, start_row, visible_rows, start_cell, visible_cells):
        emu = self.virtual_linked_base.emulator
        self.visible_history_start_row = start_row
        self.parsed = emu.calc_stringified_history(start_row, visible_rows)

    @property
    def needs_rebuild(self):
        v = self.virtual_linked_base
        emu = v.emulator
        return not self.current_num_rows == emu.num_cpu_history_entries

    def rebuild(self):
        v = self.virtual_linked_base
        emu = v.emulator
        self.current_num_rows = len(emu.cpu_history)
        segment = DefaultSegment(emu.cpu_history.entries.view(np.uint8))
        v.segment = segment
        print("CPU HISTORY ENTRIES", self.current_num_rows)
        self.init_boundaries()


class InstructionHistoryGridControl(SegmentGridControl):
    default_table_cls = InstructionHistoryTable

    def calc_default_table(self, linked_base):
        table = self.default_table_cls(linked_base)
        table.rebuild()  # find number of rows so scrollbars can be set properly
        return table

    def calc_line_renderer(self):
        return cg.VirtualTableLineRenderer(self, 2, widths=self.default_table_cls.column_sizes, col_labels=self.default_table_cls.column_labels)

    def move_viewport_to_bottom(self):
        self.move_viewport_origin((self.table.current_num_rows, 0))

    def recalc_view(self):
        log.debug(f"recalc_view: {self}")
        self.table.rebuild()
        cg.CompactGrid.recalc_view(self)
        self.move_viewport_to_bottom()

    def refresh_view(self):
        if self.IsShown():
            log.debug("refreshing %s" % self)
            if self.table.needs_rebuild:
                self.recalc_view()
            else:
                SegmentGridControl.refresh_view(self)
        else:
            log.debug("skipping refresh of hidden %s" % self)


class InstructionHistoryViewer(SegmentViewer):
    name = "cpuhistory"

    ui_name = "Instruction History"

    viewer_category = "Emulator"

    control_cls = InstructionHistoryGridControl

    # trait defaults

    # initialization

    @classmethod
    def replace_linked_base(cls, linked_base):
        # the new linked base decouples the cursor here from the other segments
        segment = DefaultSegment(np.arange(400, dtype=np.uint8))
        return VirtualTableLinkedBase(editor=linked_base.editor, segment=segment)

    def create_post(self):
        self.linked_base.table = self.control.table

    # properties

    @property
    def table(self):
        return self.control.table

    # @on_trait_change('linked_base.editor.document.byte_values_changed')
    def byte_values_changed(self, index_range):
        log.debug("byte_values_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.table.rebuild()

    # @on_trait_change('linked_base.editor.document.byte_style_changed')
    def byte_style_changed(self, index_range):
        log.debug("byte_style_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.table.rebuild()

    def recalc_data_model(self):
        self.table.rebuild()

    def do_priority_level_refresh(self):
        self.control.recalc_view()
        self.refresh_view(True)



import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)

from .emulator import known_emulators, default_emulator

from . import debugger
from .errors import *


def find_emulator(emulator_name):
    for e in known_emulators:
        if e.name == emulator_name or e == emulator_name:
            return e
    raise UnknownEmulatorError("Unknown emulator '%s'" % emulator_name)


def guess_emulator(document):
    for e in known_emulators:
        if e.guess_from_document(document):
            return e
    raise UnknownEmulatorError(f"No emulator for {document.metadata.mime}")



import os
import sys
from collections import namedtuple

import numpy as np
import wx

from ..ui.screen import BitmapScreen

from sawx.ui import compactgrid as cg
from sawx.utils.command import DisplayFlags
from ..ui import segment_grid as sg
from ..viewer import SegmentViewer
from .info import VirtualTableInfoViewer
from ..editors.linked_base import VirtualTableLinkedBase

import logging
log = logging.getLogger(__name__)


class EmulatorViewer(SegmentViewer):
    viewer_category = "Emulator"

    has_caret = False

    # effectively don't allow emulator screen viewers to use the priority
    # refresh; they will always be explicitly refreshed with the
    # emulator_update_screen_event
    priority_refresh_frame_count = 10000000

    @property
    def emulator(self):
        return self.linked_base.editor.document.emulator

    @property
    def linked_base_segment_identifier(self):
        return ""

    def use_default_view_params(self):
        pass

    def restore_view_params(self, params):
        pass

    def update_toolbar(self):
        pass

    def recalc_data_model(self):
        # self.control.recalc_view()
        self.control.Refresh()

    def recalc_view(self):
        # self.control.recalc_view()
        self.control.Refresh()

    def update_window_title(self):
        pass

    # @on_trait_change('linked_base.editor.document.emulator_update_screen_event')
    def process_emulator_update_screen(self, evt):
        log.debug("process_emulator_update_screen for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        if evt is not Undefined:
            self.control.show_frame(force=True)
            self.update_window_title()


class VideoViewer(EmulatorViewer):
    name = "video"

    ui_name = "Emulator Video Output"

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        return BitmapScreen(parent, linked_base.emulator)

    def show_caret(self, control, index, bit):
        pass

    @property
    def window_title(self):
        emu = self.linked_base.emulator
        return f"{emu.ui_name} (frame {emu.current_frame_number})"

    def update_window_title(self):
        self.update_caption()
        # FIXME: probably shouldn't know this much about the internals to call
        # blah.blah.title_bar
        self.control.GetParent().title_bar.Refresh()

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0


class CPU6502Table(sg.SegmentVirtualTable):
    want_col_header = True
    want_row_header = False

    col_labels = ["A", "X", "Y", "N", "V", "-", "B", "D", "I", "Z", "C", "SP", "PC"]
    col_sizes = [2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 4]
    p_bit = [0, 0, 0, 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01, 0]

    # cpu state is not necessarily in "A,X,Y,SP..." order, so use the dtype
    # names of the cpu state array to look up which positions they should be
    # mapped to. The array below is the default from atari800 which is
    # a, p, sp, x, y, _, pc. It's overridden below in compute_col_lookup
    col_from_cpu_state = [0, 3, 4, 1, 1, 1, 1, 1, 1, 1, 1, 2, 6]

    def calc_last_valid_index(self):
        return self.items_per_row

    def get_data_style_view(self, linked_base):
        data = linked_base.emulator.cpu_state
        self.compute_col_lookup(data.dtype)
        return data, data

    def compute_col_lookup(self, dtype):
        print(dtype)
        label_lookup = {label:i for i,label in enumerate(dtype.names)}
        col_from_cpu_state = [0] * len(self.col_labels)
        for i, d in enumerate(self.col_labels):
            if d in "NV-BDIZC":
                d = "P"
            if d in label_lookup:
                col_from_cpu_state[i] = label_lookup[d]
        self.col_from_cpu_state = col_from_cpu_state

    def get_value_style(self, row, col):
        val = self.data[self.col_from_cpu_state[col]]
        if col < 3 or col == 11:
            text = "%02x" % val
        elif col < 11:
            text = self.col_labels[col] if val & self.p_bit[col] else "-"
        else:
            text = "%04x" % val
        return text, 0

    def get_label_at_index(self, index):
        return "cpu"


class CPUParamTableViewer(VirtualTableInfoViewer):
    name = "<base class>"

    viewer_category = "Emulator"

    ui_name = "<pretty name>"

    control_cls = sg.SegmentVirtualGridControl

    @property
    def linked_base_segment_identifier(self):
        return ""


class CPU6502Viewer(CPUParamTableViewer):
    name = "cpu6502"

    ui_name = "6502 CPU Registers"

    override_table_cls = CPU6502Table


class DtypeTable(sg.SegmentVirtualTable):
    emulator_dtype_name = "<from subclass>"
    col_labels = ["hex", "dec", "bin"]
    col_sizes = [4, 4, 8]
    want_col_header = False
    want_row_header = True
    verbose = False

    def calc_num_rows(self):
        self.row_labels = []
        offsets = []
        self.label_char_width = 1
        for i, label in enumerate(self.data.dtype.names):
            if label.startswith("_") and not self.verbose:
                continue
            self.row_labels.append(label)
            offsets.append(i)
            size = len(label)
            if size > self.label_char_width:
                self.label_char_width = size
        self.row_offsets_into_data = np.asarray(offsets, dtype=np.int32)
        return len(self.row_labels)

    def calc_last_valid_index(self):
        return (self.items_per_row * self.num_rows) - 1

    def get_data_style_view(self, linked_base):
        data = linked_base.emulator.calc_dtype_data(self.emulator_dtype_name)
        print((data.dtype.names))
        return data, data

    def create_row_labels(self):
        # already generated above
        pass

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        for line in range(start_line, last_line, step):
            yield self.row_labels[line]

    def calc_row_label_width(self, view_params):
        return max([view_params.calc_text_width(r) for r in self.row_labels])

    def get_value_style(self, row, col):
        val = self.data[self.row_offsets_into_data[row]]
        try:
            if col == 0:
                text = f"{val:02x}"
            elif col == 1:
                text = f"{val}"
            elif col == 2:
                text = f"{val:08b}"
        except TypeError:
            text = "..."
        return text, 0

    def get_label_at_index(self, index):
        return self.emulator_dtype_name


# This dynamically creates clases based on the emulator-defined dtypes.
#
# Each resulting class looks like this ANTIC example, except for everywhere
# that ANTIC appears, the other class name is substituted. Much repetition! So,
# instead, we create the classes in a loop.
#
# class ANTICTableViewer(CPUParamTableViewer):
#     name = "antic"
#     ui_name = "ANTIC Registers"
#     override_table_cls = type('ANTICTable', (DtypeTable,), {'emulator_dtype_name': 'ANTIC'})

for dtype_name in ['ANTIC', 'GTIA', 'POKEY', 'PIA']:
    clsname = '%sViewer' % dtype_name
    cls = type(clsname, (CPUParamTableViewer,), {
        'name': dtype_name.lower(),
        'ui_name': '%s Registers' % dtype_name,
        'override_table_cls': type('%sTable' % dtype_name, (DtypeTable,), {'emulator_dtype_name': dtype_name}),
        })
    globals()[clsname] = cls


from .disasm import labels1

class LabelTable(cg.VariableWidthHexTable):
    want_col_header = False
    want_row_header = True

    def __init__(self, linked_base):
        self.linked_base = linked_base

        s = linked_base.segment
        super().__init__(s.data, s.style, [], s.origin)
        self.rebuild()

    def size_of_entry(self, d):
        return d[2]

    def parse_table_description(self, desc):
        row_to_label_number = []
        items_per_row = []
        index_of_row = []
        row_of_index = []
        index = 0
        row = 0
        i = 0
        print(desc)
        try:
            self.row_to_label_number = desc.keys()
        except AttributeError:
            self.row_to_label_number = []
        for i in self.row_to_label_number:
            d = desc[i]
            s = self.size_of_entry(d)
            items_per_row.append(s)
            index_of_row.append(index)
            row_of_index.extend([row] * s)
            print(f"row_of_index: {d[0]} index={index}, s={s} {d}") #: {row_of_index}")
            # print(f"index_of_row: {index_of_row}")
            index += s
            row += 1
        self.index_of_row = index_of_row
        self.row_of_index = row_of_index
        self.items_per_row = items_per_row
        self.last_valid_index = index

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        for line in range(start_line, last_line, step):
            yield self.labels[self.row_to_label_number[line]][0]

    def calc_row_label_width(self, view_params):
        return max([view_params.calc_text_width(r) for r in self.get_row_label_text(0, self.num_rows)])

    def calc_last_valid_index(self):
        pass  # calculated in init_table_description

    type_code_fmt = {
        0x00: "02x",  # hex byte
        0x01: "04x",  # hex word
        0x03: "08x",  # hex long
        0x30: "03d",  # dec byte
        0x31: "05d",  # dec word
        0x33: "010d",  # dec long
        0x40: "08b",  # bin byte
        0x41: "016b",  # bin word
        0x43: "032b",  # bin long
    }

    dtype_fmt = [
        None,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint32,
    ]

    def calc_cell_widths(self):
        cell_widths = [1] * self.num_rows
        for row in range(self.num_rows):
            label_num = self.row_to_label_number[row]
            type_code = self.labels[label_num][3]
            cell_widths[row] = len(format(0, self.type_code_fmt[type_code])) // 2  # each cell is 2 chars wide
        # print(cell_widths)
        return cell_widths

    def get_value_style(self, row, col):
        label_num = self.row_to_label_number[row]
        type_code = self.labels[label_num][3]
        bytes_per_col = (type_code & 0x03) + 1
        index = label_num + (col * bytes_per_col)
        dtype = self.dtype_fmt[bytes_per_col]
        try:
            value = int(self.data[index:index + bytes_per_col].view(dtype))
        except IndexError:
            text = ""
            style = 0
        else:
            style = self.style[index]
            text = format(value, self.type_code_fmt[type_code])
        return text, style

    def get_label_at_index(self, index):
        row = self.row_of_index[index]
        return self.labels[self.row_to_label_number[row]][0]

    def prepare_for_drawing(self, start_row, visible_rows, start_cell, visible_cells):
        # for i, item in enumerate(self.labels):
        #     print(i, item)
        pass

    def rebuild(self):
        self.labels = labels1.labels
        self.init_table_description(self.labels)
        self.init_boundaries()
        print(f"new num_rows: {self.num_rows}")


class LabelGridControl(sg.SegmentGridControl):
    default_table_cls = LabelTable
 
    def calc_default_table(self, linked_base):
        table = self.default_table_cls(linked_base)
        self.items_per_row = table.items_per_row
        self.want_row_header = table.want_row_header
        self.want_col_header = table.want_col_header
        return table

    def calc_line_renderer(self):
        print(self.table.items_per_row)
        return cg.VariableWidthLineRenderer(self, 2, self.table.items_per_row, self.table.calc_cell_widths())

    def recalc_view(self):
        log.debug(f"recalc_view: {self}")
        self.table.rebuild()
        self.line_renderer = self.calc_line_renderer()
        super().recalc_view()



class LabelViewer(SegmentViewer):
    name = "labels"

    viewer_category = "Emulator"

    ui_name = "Labels"

    control_cls = LabelGridControl

    @property
    def linked_base_segment_identifier(self):
        return ""



import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)

known_assemblers = []
default_assembler = None

try:
    from .pyatasm import ATasm
    known_assemblers.append(ATasm)
    if default_assembler is None:
        default_assembler = ATasm
except ImportError as e:
    log.warning(f"atasm (MAC/65) assembler not available: {e}")

from .. import errors
from .assembler_base import AssemblerBase, AssemblerResult

def find_assembler(assembler_name):
    for e in known_assemblers:
        if e.name == assembler_name or e == assembler_name:
            return e
    raise errors.UnknownAssemblerError("Unknown assembler '%s'" % assembler_name)



# Standard library imports.
import sys
import os

# Major package imports.
import wx
import numpy as np
import json

# Enthought library imports.
from traits.api import Any, Bool, Int, Str, List, Dict, Event, Enum, Instance, File, Unicode, Property, provides, on_trait_change
from pyface.key_pressed_event import KeyPressedEvent

# Local imports.
from sawx.framework.editor import FrameworkEditor
import sawx.framework.clipboard as clipboard
from sawx.utils.file_guess import FileMetadata
from sawx.ui.tilemanager import TileManager
from sawx.persistence import get_template
from ..byte_edit.byte_editor import ByteEditor
from ..arch.machine import Machine, Atari800
from .document import EmulationDocument
from ..utils.segmentutil import SegmentData, DefaultSegment, AnticFontSegment
from .. import guess_emulator

from sawx.utils.processutil import run_detach

import logging
log = logging.getLogger(__name__)


class EmulatorEditor(ByteEditor):
    """Editor that holds an emulator instance and the associated viewers.
    """

    default_viewers = "hex,video,char,disasm"

    #### trait default values

    #### trait property getters

    # Convenience functions

    @property
    def emulator(self):
        return self.document.emulator

    #### ByteEditor interface

    def prepare_for_destroy(self):
        self.document.stop_timer()
        ByteEditor.prepare_for_destroy(self)

    def preprocess_document(self, doc):
        args = {}
        skip = 0
        log.debug(f"preprocess_document: EmulatorEditor, {doc}")
        if self.task_arguments:
            for arg in self.task_arguments.split(","):
                if "=" in arg:
                    arg, v = arg.split("=", 1)
                else:
                    v = None
                args[arg] = v
            if "skip_frames" in args:
                skip = int(args["skip_frames"])
        try:
            doc.emulator_type
        except:
            try:
                emulator_type = args['machine']
            except KeyError:
                emulator_type = guess_emulator(doc)
            doc = EmulationDocument.create_document(source_document=doc, emulator_type=emulator_type, skip_frames_on_boot=skip)
        doc.boot()
        log.debug(f"Using emulator {doc.emulator_type}")
        return doc



"""Segment actions
"""
import os
import sys

import wx

from sawx.action import SawxAction
from sawx.ui.minibuffer import NextPrevTextMinibuffer

from .. import commands
from ... import errors

import logging
log = logging.getLogger(__name__)


class find(SawxAction):
    def calc_name(self, action_key):
        return "Find"

    def get_minibuffer(self, action_key):
        e = self.editor
        control = NextPrevTextMinibuffer(e, commands.FindAllCommand, commands.FindNextCommand, commands.FindPrevCommand, initial=e.last_search_settings["find"])
        return control

    def perform(self, action_key):
        minibuffer = self.get_minibuffer(action_key)
        self.editor.control.show_minibuffer(minibuffer)



class AssemblerResult:
    def __init__(self):
        self.errors = []
        self.segments = []
        self.transitory_equates = {}
        self.equates = {}
        self.labels = {}


class AssemblerBase:
    name = "<base>"
    ui_name = "<pretty name>"
    cpu = "<cpu>"

    comment_char = ";"
    origin = "*="
    data_byte = ".byte"
    data_byte_prefix = "$"
    data_byte_separator = ", "

    def __init__(self, verbose=False):
        self.verbose = verbose

    def assemble(self, source):
        return AssemblerResult()

